(function (factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['models/Tab','Transmit','utf8','student'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('models/Tab'),
        require('Transmit'),
        require('utf8'),
        require('student'),
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
      // var NSLibrary = NSLibrary || {};
      // NSLibrary.Client = NSLibrary.Client || {};
      // NSLibrary.Client.Chat = NSLibrary.Client.Chat || {};
  }
}(function(Tab, Transmit,UTF8, CSParams) {
    var bgPage = chrome.extension.getBackgroundPage();

    chats = {
      instance: undefined
    };

    class chat {
        constructor(chatID, CID) {
            if (chatID) { this.chatID = chatID } 
            else { this.chatID = -1; }
            this.CID = CID;
            this.target=false;
            this.isChatting =0;
            this.lastTab = false;
            this.cachedMessages = [];
            this.title = '';
            this.members = [];
            this.removeURL = true;
            this.processMessage = true;
        }

        startChat(){
            //open window
            chat.members = [];
            chat.cachedMessages = [];
            this.target = false;
            // var self = this;
            this.makeTab();
            // chrome.tabs.query({ currentWindow: true, active: true }, t => {
            //   // if (t && t[0]){
            //   //   chat.lastTab = t[0].id;
            //   // } else {
            //   //   chat.lastTab = false;
            //   // }
            //   this.makeTab();
            //   chat.makeTab();
            // });
        }

        makeTab() {
          Tab.Make('chat.html', 'Chat')
              .then(tabInfo => {
                  this.isChatting = 1;
                  this.chatTimer = window.setInterval(() => {
                      if (!CSParams.TabStore[tabInfo.key]) {
                          this.eject();
                          this.endChat();
                      }
                  }, 1000);

                  chat.chatPageStatusTimer = window.setInterval(() => {
                      chrome.tabs.query({ active: true }, tabs => {
                          var chatUrl = bgPage.chatUrl;
                          if (tabs[0] !== undefined && tabs[0].url === chatUrl) {
                              this.processMessage = true;
                              this.removeURL = false;
                          } else {
                              chrome.history.search({ text: chatUrl, maxResults: 1 }, historyItems => {
                                  if (historyItems.length == 1) {
                                      chrome.tabs.query({ active: false }, tabs => {
                                          var urlHidden = true;
                                          for (let tabInfo of tabs) {
                                              if (tabInfo.url === chatUrl) {
                                                  urlHidden = false;
                                                  break;
                                              }
                                          }

                                          this.processMessage = false;

                                          if (urlHidden) {
                                              this.removeURL = true;
                                          } else {
                                              this.removeURL = false;
                                          }
                                      });
                                  }
                              });
                          }
                      });
                  }, 500);
              });
        }

        sendChatCommand(cmd, text) {
          var chatMsg = String.fromCharCode(130, 0) +  this.chatID + String.fromCharCode(cmd, 0, 0, 0);
          chatMsg += "Complex Name";

          while (chatMsg.length < 26) chatMsg += z;

          chatMsg += UTF8.Encode(text, this.CID) + z;
          chatMsg += UTF8.Encode(localStorage.clientName, this.CID) + z;

          console.log ("==> %s %s %s", this.CID, chatMsg, text);
          new Transmit(this.CID, chatMsg).Send(chatMsg);

        }

        endChat(){
          if (this.CID)
          {
            this.sendChatCommand(3, "");
          }

          this.title = "";
          this.isChatting=0;
          Tab.Remove('Chat');
        }

        sendChatJoin() {
          this.sendChatCommand(2, "⠀");
        }

        addMessage(from, message) {
          if (this.processMessage) {
            this.createMessageElement(from, message);
          } else {
            if (this.removeURL) {
              chrome.history.deleteUrl({ url: chatUrl }, this.onUrlDeleted(from, message));
            } else {
              this.eject();
              this.createTabWithMessage(from, message);
            }
            return;
          }
        }

        addJoinedMessage(from){
          if (this.members.indexOf(from)<0){
            var span = document.createElement("div");
            span.innerHTML = "<b>" + NSLibrary.Translate('PLACEHOLDER_HAS_JOINED', from) + "</b>";
            span.classList.add('join');
            this.cachedMessages.push(span);
            if (this.target){
              this.target.appendChild(span);
              this.target.scrollTop = 99999999;
            }

            this.members.push(from);
          }
        }

        addLeftMessage(from){
          const translate = NSLibrary.Translate('PLACEHOLDER_HAS_LEFT', from);

          const span = document.createElement("div");
          span.innerHTML = `<b>${translate}</b>`;
          span.classList.add('leave');
          chat.cachedMessages.push(span);
          if (this.target){
            this.target.appendChild(span);
            this.target.scrollTop = 99999999;
          }
          let mIX = this.members.indexOf(from);
          if (mIX>-1)
          {
            // Remove item at mIX;
            this.members.splice(mIX,1);
          }

          // Remove item at mIX;
          //if (mIX + 1) this.members.splice(mIX,1);
        }

        sendMessage(message){
          if (message.replace(/\s/g,'').length > 0){
            this.sendChatCommand(4, message);
            this.addMessage(localStorage.clientName, message);
          }
        }

        setTarget(element){
          var c = chats.instance;
          c.target = element;
          //check if we have any cached messages
          var length = c.cachedMessages.length;
          for (var i=0;i<length;i++){
            c.target.appendChild(c.cachedMessages[i]);
          }
          element.previousElementSibling.innerHTML = c.title;
          c.target.scrollTop = 99999999;
        }

        eject() {
          var c = chats.instance;
          if (c.removeURL && bgPage && bgPage.chatUrl) {
            chrome.history.deleteUrl({ url: bgPage.chatUrl }, this.onUrlRemoved);
          } else {
            window.clearInterval(this.chatTimer);
            window.clearInterval(this.chatPageStatusTimer);
            Tab.Remove('Chat');
          }
        }

        setTitle(title){
          if (!title || title===""){
            this.title = "";
          }else{
            this.title = NSLibrary.Translate('TOPIC') +": " + title;
          }
        }

        /**
          Utility Methods
         */

        createTabWithMessage(from, message) {
          this.makeTab();
          this.createMessageElement(from, message);
        }

        createMessageElement(from, message) {
            var span = document.createElement("div");
            span.innerHTML = "<b>" + from + ' ' + NSLibrary.Translate('SAYS') + ": </b><br/>&nbsp;&nbsp;&nbsp;" + chat.detectLinks(message.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;'));
            span.classList.add('message');
            this.cachedMessages.push(span);
            if (this.target) {
                this.target.appendChild(span);
                this.target.scrollTop = 99999999;
            }
        }

        updateTabStoreAndCreateTabWithMessage(from, message) {
          window.clearInterval(this.chatPageStatusTimer);
          window.clearInterval(this.chatTimer);
          chat.clearTabID('Chat');
          this.createTabWithMessage(from, message);
        }

        /**
        Events
         */
        onUrlDeleted(from, message) {
          this.updateTabStoreAndCreateTabWithMessage(from, message);
        }

        onUrlRemoved() {
          window.clearInterval(this.chatPageStatusTimer);
          window.clearInterval(this.chatTimer);
          chat.clearTabID('Chat');
        }

        static clearTabID(name) {
          var tabId = CSParams.TabStore[name];
          if (tabId) {
            Tab.clearId(tabId)
          }
        }

        static detectLinks(text) {
            var urlRegex = /(\b(https?|ftp|file):\/\/[-A-Z0-9+&@#\/%?=~_|!:,.;]*[-A-Z0-9+&@#\/%=~_|])/ig;
            return text.replace(urlRegex, function(url) {
                return '<a href="' + url + '">' + url + '</a>';
            });
        }

        static instance(value){
          if (value) {
            chats.instance = value;
          } else {
            return chats.instance;
          }
        }
    }


    return chat;

}));
