(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['modules/debug'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('modules/debug')
      );
  } else {
      // Browser globals (root is window)
      // var module = module || null;
      // var NSLibrary = NSLibrary || {};
      // root.NSLibrary.Comms = factory();
  }
}(this, function(debug) {
  var logging = 0;

  const interfaceMethods = [
    "OPEN",
    "HEARTBEAT",
    "POLL",
    "STATUS",
    "HANGUP",
    "CONNECT_REPLY",
    "BROADCASTDATA",
    "CLOSE"
  ];

  class comms {
    constructor(mechanism) {
      this.mechanism = mechanism; // Undefined
    }

    /**
     * SWITCHING MECHANISM to activate replacement comms provider.
     */


    activate(communicator) {
      debug("Activating Comms");

      try {
        interfaceMethods.map(m => {
          // var handle = interfaceMethods[m];
          if (communicator[m]) {
            debug(m + " not found in supplied method");
            this[m] = communicator[m];
          }
        });
      } catch (e) {
        debug(e);
      }
    }

    /**
     * INTERFACE METHODS
     */


    OPEN(){}
    HEARTBEAT(){}
    POLL(){}
    STATUS(){}
    HANGUP(){}
    CONNECT_REPLY(){}
    BROADCASTDATA(){}
    CLOSE(){}
  }

  comms.debug = setDebugger;
  // comms.WebSockets = { active: undefined, expired: [] };


  function setDebugger(show) {
    logging = show ? true : false;
  }

/**
 * THIS IS NOT NEEDED WHEN USING ES6 CLASS Implementation.
 */

  // function initialize(value) {
  //   // initializes a hash with @value
  //   // returns Object (shallow copy)
  //   const methods = [...interfaceMethods];
  //   const object = {};
  //   methods.forEach(key => (object[key] = value));
  //   return Object.assign({}, object);
  // }

  // initialize the active and send methods
  // comms.activate(comms.active);



  comms.str2ab = str2ab;

  function str2ab(str) {
      if (!str || str.constructor !== String) return;
      var buf = new ArrayBuffer(str.length);
      var bufView = new Uint8Array(buf);
      for (var i = 0, strLen = str.length; i < strLen; i++) {
          bufView[i] = str.charCodeAt(i);
      }
      return buf;
  }

  return comms;
}));
