(function (root, factory) {

  if (typeof define === 'function' && define.amd) {
      // AMD
      define([
        'comms.nc.protocol',
        'utf8',
        'flag',
        'modules/translate',
        'pubsub',
        'student',
        'background/CLIENT',
        'accept',
        'mouse',
        'chatlib',
        'showlib',
        'models/register',
        'models/Tab',
        'models/survey',
        'modules/image',
        'modules/extension',
        'background/subscription',
        'background/thumb/videoStream',
        'models/sysinfo',
        'modules/debug'
      ], factory);
  } else if (typeof exports === 'object') {
      //
      // Node, CommonJS-like
      module.exports = factory(
          require('comms.nc.protocol'),
          require('utf8'),
          require('flag'),
          require('background/translate'),
          require('pubsub'),
          require('student'),
          require('background/CLIENT'),
          require('accept'),
          require('mouse'),
          require('chatlib'),
          require('models/survey'),
          require('modules/image'),
          require('modules/extension'),
          require('background/subscription'),
          require('background/thumb/videoStream'),
          require('models/sysinfo'),
          require('modules/debug')
        );
  } else {
      // Browser globals (root is window)
      // import CSParams from "./student.js";
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);

      // UTF8 = root.NSLibrary.UTF8;
      // // NSLibrary.Comms = root.NSLibrary.Comms || { };
      // NSLibrary.Comms.NC = factory.apply(NSLibrary.Comms.NC,
      //   NSLibrary.Comms.NC.protocol,
      //   root.NSLibrary.UTF8,
      //   require('flag'),
      //   require('pubsub'),
      //   csParams,
      //   root.NSLibrary.Client
      // );
  }
}(this, function(protocol, UTF8, FLAG, TRANSLATE, PUBSUB, CSPARAMS, CLIENT, getNCAccept, Mouse, Chat, Show, Register, Tab, Survey, Image, extension, Subscription, Stream, sysinfo, debug) {
  const nc = {};
  const NC_ACCEPT = getNCAccept();
  var showFeedBack = false;
  var registry = {};
  big = []; // for assembling big packets. each item is (0:cid, 1:id, 2:str, 3:tot)

  /*
      Private Callback Stack
      Each registered callback will recieve the
      entire NC data payload

      The registry structure is:
      Key (NC code), Callback (Function)

      Usage:
      NSLibrary.NC.Subscribe("235", Function); // eg. WebRtc Module
      NSLibrary.NC.Subscribe("235", Function); // eg. OtherJson Module

      This will create an NC keyed entry
      {"235": [ Function, Function ]}

  */

  nc.Subscribe = function(NC, callback) {

    // Given @NC and @callback Function
    if (NC && callback && callback instanceof Function) {
      NC = String(NC);
      registry[NC] = registry[NC] || [];
      registry[NC].push(callback);
      console.log("Added NC subscriber for [%s]", NC);
      return;
    }

    // Otherwise if given only @NC ...
    return registry[NC] || null;
  };
  
  function contents(source)
  {
    let ab = new ArrayBuffer(source.length*4);
    let view = new Uint32Array(ab);
    for (let a = 0;a<source.length; a++)
    {
      view[a] = source.charCodeAt(a);
    }
    return view;
  }
  
  nc.Process = function(d) {
      //var nc; //! unused variable???
      this.csParams = CSPARAMS;

      this.d = d;
      this.Client = CLIENT;

      // console.log("========", contents(d));

      // Flag Evaluations
      this.FF = FLAG.Get(d,2);
      this.FFFF = FLAG.Get(d,4);

      // ncdata
      this.cid = d.substr(0, 4);
      this.nc = d.charCodeAt(4 + 0);

    //   console.log("<< NC:(%s) cid:", this.nc, this.cid.split("").map(x => x.charCodeAt(0)));

    //   console.log(this.nc);
      if (this.nc === protocol.BIGNEASY)
      {
          this.BigEasy();
          // BigEasy changes value of d
          this.FF = FLAG.Get(this.d,2);
          this.FFFF = FLAG.Get(this.d,4);
        //   console.log("BIGNEASY:", contents(d));
      }

      sm = 0;
      this.ActiveConnection = this.csParams.Connections[this.cid];
      sm = (this.ActiveConnection) ? this.ActiveConnection.sm : 0;

      if (this.nc == protocol.LOGIN && sm == 1) // nc_login
      {
          this.Login();
      }
      else if (this.nc == protocol.SELECT && sm == 2) // nc_select
      {
          this.Select([]);
      }

      this.Perform(this.nc);
  };

  nc.Perform = function(packet) {

      // NC to Process
      // packet = NC String;
      // NOTE: this = augmented properties

      // Notify the subscriber stack, if a callback exists
      // execute the callback and skip the switch.

      if (packet && registry[packet]) {
          console.log("Notifing registry for NC protocol [%s]", packet);
          const that = this;
          registry[packet].forEach(function(fn) { return fn.call(that, that.d); });
          return;
      }

      switch(packet)
      {
          case protocol.FLUSH:
              this.Flush();
              break;
          case protocol.SYSINFO:
              this.SysInfo();
              this.SendTabData();
              break;
          case protocol.RECONNECT:
              this.Reconnect();
              break;
          case protocol.TICKLE:
          case protocol.EXEC:
          case protocol.BATTERY:
          case protocol.WIRELESS:
              // less noise if we just ignore these nsm nc packets
              break;

          case protocol.DCAFPAL:
              // NSLibrary.Comms.NC.Palette();
            //   console.debug('DEBUG REPLACED METHOD HERE');
              this.Palette();
              break;
          case protocol.SDATA:
              this.ActiveConnection.stream += this.d.substr(6);
              break;
          case protocol.EDATA:
              alert('gavin – fix edata');
              this.ActiveConnection.stream = '';
              break;
          case protocol.SOS_CLEAR:
              // Cancel Notification
              if (gHRNotification){
                  gHRNotification.close();
              }

              break;

          case protocol.LATLONG:
              // Manager Feature.
              //getLatLng(this.cid);
              break;
          case protocol.ECDATA:
              break;
          case protocol.CLIPDATA:
              this.Clipboard();
              break;

          case protocol.SETMODE:
              var mode = this.FF(6);
              this.SetMode(mode);
              break;
          case protocol.NEWSCRAPE:
              this.NewScrape();
              break;
          case protocol.WSLAVE:
              this.ShowScaling();
              break;
          case protocol.VIDEOHINT:
              this.VideoHint();
              break;
          case protocol.GETMONITORS:
              CSPARAMS.get.currentWindow()
              .then(Stream.getDimensions)
              .then(w => {
                // console.log('GETMONITOR PROTOCOL (STREAM):', w.width, w.height);
                if (w.width){
                    var getmon = function(w,h){ return String.fromCharCode(protocol.GETMONITORS) + z + z2 + String.fromCharCode(1) + z + z2 + z4 + z4 + z4 + String.fromCharCode(w & 255) + String.fromCharCode(w >> 8) + z2 + String.fromCharCode(h & 255) + String.fromCharCode(h >> 8) + z2 + z; };
                    nc.ChromeWindow(CLIENT.View, w, 0, getmon);
                }
              });

              break;
          case protocol.KEYSTROKES:
               // keyboard
              this.Keyboard();
              break;
          case protocol.GIVE_REWARD:
          case protocol.AUDIOSTAT:
          case protocol.JOURNAL:
              // less noise if we just ignore these nss nc packets
              break;

          case protocol.STUDENTUI:
              this.StudentUI();
              break;

          case protocol.MESSAGE:
              this.Message();
              break;

          case protocol.VOTE:
              this.Vote();
              break;

          case protocol.VOTESHOWRESULT:
              var bgPage = chrome.extension.getBackgroundPage();
              if (bgPage.surveyAnswered) {
                  this.VoteShowResult();
              }
              break;

          case protocol.CHATEX:
              this.Chat();
              break;
          case protocol.APPROVEDWEBLIST:
              this.ApprovedWebList();
              break;
          case protocol.WINMOUSE:
              this.WinMouse();
              break;
          case protocol.THUMBNAIL:
              this.Thumbnail();
              break;
          case protocol.PLUGIN:
              // console.debug('REPLACED Plugin packet');
              this.Plugin();
              break;
          case protocol.REGISTER_USER:
              CSParams.Register = new Register(this.cid);
              this.RegisterStudent(this.d);
            break;
          case protocol.CANCEL_REGISTER:
              extension.cancelRegister();
              break;

          case protocol.NC_JSON: //235
              console.log("Got NC_JSON for (%s)", this.cid);
              this.ncJson();
              break;

          default:
              // console.warn("UNKNOWN protocol", packet, this.d, this.cid);
              if(this.d.length >= 4) {
                  // some other nc packet, so display info about it
                  //var nclen = d.length - 4;
                  //nfr alert("nc=" + packet + ", nclen=" + nclen);
              }
      }
  };

  nc.ncJson = function (message) {
      let obj;

      // if @message is defined it was
      // intercepted from nc.Message()
      if(message) {
          try {
              obj = JSON.parse(message);
              if(!obj || !obj.key) return false;
          } catch (e) {
              return false;
          }
      }

      // Delegate data to WebRTC module
      let {cid} = this;
      setTimeout(() => {
          if(isNaN(Number(cid)) && cid.length === 4) {
              cid = cid.split("").reverse().map(c => c.charCodeAt(0));
              cid = bytes(cid, 4);
          }

          // console.log("Delgate NS_JSON to NSLibrary.Comms.webrtc;", cid, obj);
          const that = this;
          registry[String(235)].forEach(function(fn) { return fn.call(that, {cid, Data: message}); });
      }, 0);

      return true;
  };


  nc.Login = function() {

    flags = this.FF(6)
      if (flags & 0x200) {
          //IS pin conect
          var pin = this.d.substr(6+35, 5);
          var reject;
          if (pin != gPin){

              reject = String.fromCharCode(protocol.REJECT, 0, 25, 0, 0); // notmypin

          }else{

              reject = String.fromCharCode(protocol.REJECT, 0, 5, 0, 0);
          }
          extension.senddata(this.cid, reject);
      }else{
          // nc_login sending nc_config and nc_accept
          //var config = chr(82) + z + z2 + 'Client' + z + 'DisableAudio=9' + z + 'DisableBatteryInfo=9' + z + 'DisableAnnotate=9' + z + 'DisableBlankScreen=9' + z + 'DisableChat=9' + z + 'DisableControl=9'/* + z + 'DisableShare=9'*/ + z + 'DisableExec=9' + z + 'DisableFileTransfer=9' + z + 'DisableInventory=9' + z + 'DisableJournal=9' + z + 'DisableLogoff=9' + z + 'DisableLogon=9' + z + 'DisablePowerOff=9' + z + 'DisablePrintCapture=9' + z + 'DisableReboot=9' + z + 'DisableRemoteCmd=9' + z + 'DisableRegistryIO=9' + z + 'DisableSAS=9' + z + 'DisableStudentUI=9' + z + 'DisableWirelessInfo=9' + z + 'MacAddress=' + localStorage.macAddress + z + z;
          // var config = String.fromCharCode(protocol.CONFIG) + z + z2 + 'Client' + z + 'DisableAudio=9' + z + 'DisableBatteryInfo=9' + z + 'DisableAnnotate=9' + z + 'DisableBlankScreen=9' +/* z + 'DisableChat=9' +*/ z + 'DisableControl=9'/* + z + 'DisableShare=9'*/ + z + 'DisableExec=9' + z + 'DisableFileTransfer=9' + z + 'DisableInventory=9' + z + 'DisableJournal=9' + z + 'DisableLogoff=9' + z + 'DisableLogon=9' + z + 'DisablePowerOff=9' + z + 'DisablePrintCapture=9' + z + 'DisableReboot=9' + z + 'DisableRemoteCmd=9' + z + 'DisableRegistryIO=9' + z + 'DisableSAS=9' + z + 'DisableStudentUI=9' + z + 'DisableWirelessInfo=9' + z + 'MacAddress=' + localStorage.macAddress + z + 'Channel=' + ((UTF8.Required(localStorage.room))?' ':localStorage.room) + z + 'ChromeUtf8=1' + z + 'ChannelUtf8=' + UTF8.ForceEncode(localStorage.room) + z + '_License' + z + 'Product=' + productCode + z + z;
          var config = String.fromCharCode(protocol.CONFIG) + z + z2 +
          'Client' + z + 'DisableAudio=9' + z +
          'DisableBatteryInfo=9' + z + 'DisableAnnotate=9' + z +
          'DisableBlankScreen=9' +/* z + 'DisableChat=9' +*/ z +
          'DisableControl=9' + z + 'DisableShare=9' + z +
          'DisableExec=9' + z + 'DisableFileTransfer=9' + z +
          'DisableInventory=9' + z + 'DisableJournal=9' + z +
          'DisableLogoff=9' + z + 'DisableLogon=9' + z +
          'DisablePowerOff=9' + z + 'DisablePrintCapture=9' + z +
          'DisableReboot=9' + z + 'DisableRemoteCmd=9' + z +
          'DisableRegistryIO=9' + z + 'DisableSAS=9' + z +
          'DisableStudentUI=9' + z + 'DisableWirelessInfo=9' + z +
          'MacAddress=' + localStorage.macAddress + z +
          'Channel=' + (UTF8.Required(localStorage.room) ? ' ' : localStorage.room) + z +
          'ChromeUtf8=1' + z + 'ChannelUtf8=' + UTF8.ForceEncode(localStorage.room) +
          (UTF8.Required(localStorage.clientName) ? z : z + 'CurrentUserName=' + localStorage.clientName + z) +
          '_License' + z + 'Product=' + productCode + z + z;

          extension.senddata(this.cid, config);

          //var accept2=chr(4)+z + chr(201)+z + z2+z4+z4 + chr(3)+z + z2 + z4 + z2 + chr(127)+chr(247) + chr(1)+z + chr(255)+chr(255)+chr(42)+z + chr(20)+z + chr(32)+z + z;
          //var accept2 = chr(4) + z + chr(201) + z + z2 + z4 + z4 + chr(3) + z + z2 + z4 + z2 + chr(64) + chr(0xc5) + chr(1) + z + chr(255) + chr(255) + chr(42) + chr(1) + chr(20) + z + chr(32) + z + z;
          extension.senddata(this.cid, NC_ACCEPT);

          // we've received nc_login, and we've sent back empty nc_config and nc_accept
          // so update the state machine (expecting nc_select)
          this.ActiveConnection.sm = 2;
      }
  };

  nc.Reconnect = function() {
    this.csParams.reconnectMode.active = this.d.charCodeAt(4+2);
    this.csParams.reconnectMode.id = this.d.substr(4+35,38);
    debug('RECONNECT Received:' + this.csParams.reconnectMode.id);
  };

  nc.Select = function() {
      // Evaluate Room
      var payload = this.d.substring(4+12, this.d.length).split(z);
      var channelName = payload[8];
      var storedRoom = sessionStorage.room;
      var reject;

      if (channelName === "")
      {
          // Accept Connection
          PUBSUB.publish("connect",{cid:this.cid, d:this.d});
          return;
      }

      // TODO: RE-integrate this
      //
      // if (false/*!storedRoom*/)
      // {
      //     // No Room selected
      //     // WRONGCHANNEL
      //     reject = String.fromCharCode(protocol.REJECT, 0, 23, 0, 0);
      //     extension.senddata(this.cid, reject);
      //     return;
      // }

      if (payload.length > 10)
      {
          // UTF-8 Support
          channelName = UTF8.ForceDecode(payload[10]);
      } else {
          // Pre-UTF8
          if (channelName)
          {
            for (i=0; i<channelName.length; i++)
            {
              if (channelName.charCodeAt(i) > 127)
              {
                // Non-ASCII, Connect.
                PUBSUB.publish("connect",{cid:this.cid, d:this.d});
                return;
              }
            }

          }
      }

      if (channelName && (channelName.toLowerCase() !== sessionStorage.room.toLowerCase()))
      {
          // WRONGCHANNEL
          reject = String.fromCharCode(protocol.REJECT, 0, 23, 0, 0);
          extension.senddata(this.cid, reject);
      } else {
          // Connect
          PUBSUB.publish("connect",{cid:this.cid, d:this.d});
      }
  };

  nc.Flush = function() {
      flags = this.FF(6+6);
      if (flags & 0x1000)
      {
          this.ActiveConnection.utf8 = true;
          UTF8.SetUTF8(this.ActiveConnection.cid);
      }
  };

  nc.SysInfo = function(){
    extension.senddata(this.cid, sysinfo.string);
  };

  nc.Vote = function()
  {
      debug('in VOTE');
      var d = this.d;
      var tabIndex = d.indexOf(String.fromCharCode(9), 4 + 2);
      if (tabIndex >= 0) {
          zx = d.indexOf(z, tabIndex + 1);
          question = UTF8.Decode(d.substring(4 + 4, tabIndex), this.cid);
          options = UTF8.Decode(d.substring(tabIndex + 1, zx), this.cid);

          // Replace ';' or ', '
          options = options.trim().replace(/ *; *| *, */g, ',');
          //var aopts = options.split(',');
          //alert("survey, question=" + question + ", options=" + options);
          CSPARAMS.Survey = new Survey(this.cid, question, options);
          Tab.Make('survey.html', 'Survey');
      }
      else { //alert("survey cancelled");
          extension.cancelSurvey();
      }
  };

  nc.VoteShowResult = function()
  {
      debug('in VOTESHOWRESULT');
      var votes, total;

      var newTotal = 0;
      zx = this.d.indexOf(z, 4 + 2);
      var itemstr = UTF8.Decode(this.d.substring(4 + 2, zx - 1), this.cid);
      var action = this.d.charCodeAt(zx + 1); // 0=show, 1=hide, 2=update

      //var resptype = d.charCodeAt(zx + 5);
      var items;
      var options = CSPARAMS.Survey.Answers;
      if (action != 1) {
          // untested -
          items = itemstr.split('\t');
          //var count = items[0];
          total = items[1];
          question = items[2];
          items.splice(0, 3);
          CSPARAMS.Survey.Answers = [];
          votes = [];
          for (i = 0; i < items.length; ++i) {
              x = items[i].split('\n');
              CSPARAMS.Survey.Answers.push(x[0]);
              votes.push(x[1]);
          }
          //alert("survey result, q=" + question + ", " count + "/" + total);
      }

      if (votes) {
          for (i = 0; i < votes.length; i++) {
              newTotal = newTotal + parseInt(votes[i], 10);
          }
      }

      //alert("survey result");
      if (CSPARAMS.Survey.Answers === undefined || votes === undefined || CSPARAMS.Survey.Answers.length == 0) {
          Tab.Remove('SurveyResults');
      } else {
          // items: [0]=count, [1]=total, [2]=question, [3..8]=options and scores
          surveyResults = '';
          var optionsLength = CSPARAMS.Survey.Answers.length;
          for (i = 0; i < optionsLength; i++) {
              var pc;
              if (total === 0)
              {
                  pc = 0;
              }
              else
              {
                  pc = (votes) ? votes[i] * 100 / newTotal : 0;
              }

              surveyResults = surveyResults + pc.toFixed(2) + '%    ' + CSPARAMS.Survey.Answers[i] + '\r\n';
          }
          extension.setSurveyResults(surveyResults);

          if (CSPARAMS.TabStore.SurveyResults) {
              chrome.tabs.reload(CSPARAMS.TabStore.SurveyResults);
          } else {
              Tab.Make('results.html?question=' + CSPARAMS.Survey.Question + '&response=' + encodeURIComponent(surveyResults),'SurveyResults');
          }
      }
  };


  nc.RegisterStudent = function(d)
  {
      debug("in STUDENTREGISTER");
      flags = this.FF(6);
      // this.csParams.Register.Flags = flags;

      f = FLAG.Check(flags);
      var student = {
          FirstName:f(1),
          Surname:f(2),
          UserName:f(512),
          ClassName:f(4),
          StudentID:f(8),
          Custom1:false,
          Custom2:false
      };

      x = 6 + 4 + 2;
      zx = d.indexOf(z, x);

      c = FLAG.CheckAssign(flags, CSPARAMS.Register.Cid);
      if(c(16,  function(v){ student.Custom1 = v; }, d.substring(x, zx))) { x = zx + 1; zx = d.indexOf(z, x); }
      if(c(32,  function(v){ student.Custom2 = v; }, d.substring(x, zx))) { x = zx + 1; zx = d.indexOf(z, x); }
      if(c(64,  function(v){ CSPARAMS.Register.Teacher = v; }, d.substring(x, zx))) { x = zx + 1; zx = d.indexOf(z, x); }
      if(c(128, function(v){ CSPARAMS.Register.Lesson = v; }, d.substring(x, zx))) { x = zx + 1; zx = d.indexOf(z, x); }
      if(c(256, function(v){ CSPARAMS.Register.Room = v; }, d.substring(x, zx))) { x = zx + 1; zx = d.indexOf(z, x); }

      this.csParams.Register.Student = student;

      //alert("Register, c1=" + custom1 + ", c2=" + custom2 + ", " + "teach=" + csParams.Register.Teacher + ", lesson=" + csParams.Register.Lesson + ", room=" + csParams.Register.Room);
      if (!student.UserName || student.ClassName || student.StudentID || student.Custom1 || student.Custom2){
          if(!this.csParams.TabStore.Register) {
              Tab.Make('register.html', 'Register');
          }else{
           chrome.tabs.get(this.csParams.TabStore.Register, function(tab) {
                  if (!tab){
                      Tab.Make('register.html', 'Register');
                  }else{
                      chrome.tabs.update(this.csParams.TabStore.Register, {url: chrome.extension.getURL('register.html')});
                  }
              });
          }
      }else{
          extension.studentRegister({userName: localStorage.userName});
      }
  };

  nc.BigEasy = function() {
      var id = this.FFFF(4+6);

      more = this.d.charCodeAt(4 + 10);
      var str = this.d.substr(4 + 11);

      found = 0;
      for (i = 0; i < big.length; ++i) {
          if (big[i][0] == this.cid && big[i][1] == id) {
              found = true;
              if (!more) {
                  this.d = this.cid + big[i][2] + str;
                  this.nc = this.d.charCodeAt(4 + 0);
                  //alert("nc_bigneasy, id=" + id + ", nc=" + packet + ", nclen=" + d.length + ", totlen=" + big[i][3]);
                  big.splice(i, 1);
              }
              else {
                  big[i][2] = big[i][2] + str;
                  return;
              }
          }
      }
      if (!found) {
          var tot = this.FFFF(4 + 2);
          // var tot = (d.charCodeAt(4 + 5) << 24) + (d.charCodeAt(4 + 4) << 16) + (d.charCodeAt(4 + 3) << 8) + d.charCodeAt(4 + 2);
          big.push([this.cid, id, str, tot]);
      }
  };

  nc.Palette = function(){
      // show = this.Client.Show;
      show = Show;
      var ncolors = this.FF(6);
      more = this.d.charCodeAt(6 + 2);
      var idx = 6 + 4;

      if (show.palette.length >= 256) {
          show.palette = [];
      }
      // global palette will be built up over 4 packets, each having 64 colour entries

      for (var c = 0; c < ncolors; ++c)
      {
          var r = this.d.charCodeAt(idx + 2);
          var g = this.d.charCodeAt(idx + 1);
          var b = this.d.charCodeAt(idx);

          show.palette.push([r, g, b]);
          idx += 4;
      }
  };

  nc.Clipboard = function() {
      var cliptext, size, idx;
      flags = this.FF(6+6);
      if (flags & 0x10) { // check
          var clip = String.fromCharCode(protocol.CLIPDATA, 0, 1, 0, 0, 0, 0, 0, 0x20, 0, 0, 0, 0, 0, 0, 0);
          extension.senddata(this.cid, clip);
      }
      else if (flags & 4) {  // send
          cliptext = CLIENT.Clipboard.getClipboardText();
          size = cliptext.length + 1;
          //var clipa = String.fromCharCode( 119,0, 1,0, size&255, (size>>8) & 255,0,0, 0x03, 0, 0,0,0,0, size&255, (size>>8) & 255 ) + cliptext + z;
          //extension.senddata(this.cid, clipa);

          size = size * 2;
          var clipu = String.fromCharCode(protocol.CLIPDATA, 0, 13, 0, size & 255, (size >> 8) & 255, 0, 0, 0x03, 0, 0, 0, 0, 0, size & 255, (size >> 8) & 255);
          for (idx = 0; idx < cliptext.length; ++idx) {
              clipu += String.fromCharCode(cliptext.charCodeAt(idx) & 255, cliptext.charCodeAt(idx) >> 8);
          }
          clipu += z2;
          extension.senddata(this.cid, clipu);

          var clipdone = String.fromCharCode(protocol.CLIPDATA, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0);
          extension.senddata(this.cid, clipdone);
      }else if ((flags & 3) == 3) {
          if (this.d.charCodeAt(6) == 1) { // ascii text
              zx = this.d.indexOf(z, 6 + 14);
              cliptext = this.d.substring(6 + 14, zx);
              CLIENT.Clipboard.setClipboardText(cliptext);
          }
          else if (this.d.charCodeAt(6) == 13) { // unicode text
              idx = 6 + 14;
              cliptext = '';
              while ((ch = this.d.charCodeAt(idx) + this.d.charCodeAt(idx + 1) * 256) !== 0) {
                  cliptext += ch;
                  idx += 2;
              }
              CLIENT.Clipboard.setClipboardText(cliptext);
          }
      }
  };

  nc.SetMode = function(mode){
      var modeflags = this.FF(8);
       // console.log("in nc_setmode");
        // console.log(contents(this.d));
       var view = CLIENT.View;

      // var Show = CLIENT.Show;
      if (mode & 8) {
          if (modeflags & 4){
              Show.isFullScreen = 1;
          }else if (modeflags & 8 || modeflags & 0x4000){
              Show.isFullScreen = 0;
          }else{
              Show.isFullScreen = 1;
          }

          Show.startShow();
      }

      if (mode & 1)
      {
          view.viewCid = this.cid;
          if (!view.isScraping) {
              Stream.enable();
              // CSParams.get.currentWindow()
              Stream.getDimensions()
              .then(nc.SendWSlave)
              .then(function(View){
                  View.isScraping = 1;
                  View.getScreen();
              });
          }
      }
      else
      {
          if (view.isScraping) {
              window.clearInterval(View.timer);
              view.isViewRestart = 1;
              view.isScraping = 0;
          }

          // if (mode == 4096 || modeflags & 1){
          if (mode == 4096){
            Show.endShow();
          this.csParams.Show.Width=undefined;
          }
      }

      if ((mode & 0x800) && (tcpConnected || !(this.ActiveConnection.mode & 0x800))) {
          CLIENT.Lock.lockClient();
      } else if (!(mode & 0x800) && (tcpConnected || (this.ActiveConnection.mode & 0x800))) {
          CLIENT.Lock.unlockClient();
      }

      if (!tcpConnected){
          this.ActiveConnection.mode = mode;
      }
  };

  nc.ShowScaling = function()
  {
      var d = this.d;
      //console.log("in nc 49");
      colorbits = d.charCodeAt(6 + 2);
      // var Show = this.Client.Show;

      this.csParams.Show.Width = this.FF(6 + 6);
      this.csParams.Show.Height = this.FF(6 + 8);
      Show.xOffset = (d.length > 21) ? this.FF(6 + 12) : 0;
      Show.yOffset = (d.length > 21) ? this.FF(6 + 14) : 0;
      Show.Height = this.csParams.Show.Height;
      Show.Width = this.csParams.Show.Width;
      Show.backBufferCanvas.width = Show.Width;
      Show.backBufferCanvas.height = Show.Height;
      Show.Scale(CSParams.ScreenDimensions.height, CSParams.ScreenDimensions.width);

      Show.dimensionsSet = true;
      // Show.displayCanvas.width = this.csParams.Show.Width;
      // Show.displayCanvas.height = this.csParams.Show.Height;
      //
      //
      // Show.backBufferCanvas.width = this.csParams.Show.Width;
      // Show.backBufferCanvas.height = this.csParams.Show.Height;
  };

  nc.VideoHint = function()
  {
      var d = this.d;
      var hint = d.charCodeAt(6);
      if (hint === 0)
      {
          // var show = this.Client.Show;
          var show = Show;
          var width = this.csParams.Show.Width;
          var height = this.csParams.Show.Height;

          show.tempShowCanvas.width = width;
          show.tempShowCanvas.height = height;

          var fragments = show.fragments;
          show.fragments = [];
          for (var i = 0; i < fragments.length; ++i)
          {
              show.tempShowContext.putImageData(fragments[i].imagedata, fragments[i].xpos, fragments[i].ypos);
          }
          show.fragments = [];
          if (show.displayContext)
          {
            show.displayContext.drawImage(show.tempShowCanvas, 0, 0, width, height, 0, 0, width, height);
          }

          if(show.dimensionsSet)
          {
            show.backBufferContext.drawImage(show.tempShowCanvas, 0, 0, width, height);


            show.backBufferContext.drawImage(show.displayCanvas, 0, 0, width, height);
            show.dimensionsSet = 0;
          }
      }
  };

  nc.Keyboard = function()
  {
      var Keyboard = CLIENT.Keyboard;
      var d = this.d;
      var extended = false;
      var nkeys = d.charCodeAt(6);
      for (var kx = 0; kx < nkeys; ++kx) {
          var key = d.charCodeAt(8 + kx);
          if (key == 255) {
              ++kx;
              key = d.charCodeAt(8 + kx);
              Keyboard.sendAscii(key);
          }
          else if (key == 224) {
              extended = !extended; // used with numlock and cursor keys
          }
          else {
              if (key < 128) {
                  Keyboard.sendScanCode(key, extended);
                  extended = false;
              }else {
                  if (key == 170 || key == 182 ) {
                      Keyboard.shiftStatus = false;
                  }else if (key == 157) {
                      Keyboard.ctrlStatus = false;
                  }else if (key == 184) {
                      Keyboard.altStatus = false;
                  }
              }
          }
      }
  };

  nc.NewScrape = function(){
      // show = this.Client.Show;
      show = Show;
          //console.log("in newscrape");
      var colourBitMultiplier;
      width = this.FF(6 + 4);
      //width = 1440;
      height = this.FF(6 + 6);
      if (height >= 32768) {
          height = 65536 - height;
          show.isUpsideDown = false;
      }else{
          //data is upside down
          show.isUpsideDown = true;
      }

      //if negative

      //Setup temp canvas for drawing this piece for the show data
      //console.log("setting canvas size");
      show.tempShowCanvas.width = width;
      show.tempShowCanvas.height = height;

      colorbits = this.d.charCodeAt(6 + 8);

      //decode and RLE decode the data from the control
      var encodedBMP = this.d.substr(6 + 9);
      var decoded = show.rleDecode(encodedBMP);

      var newFix = '';

      if (colorbits !== 8) {
          colourBitMultiplier = decoded.length / 2;
      }else {
          colourBitMultiplier = decoded.length;
      }

      if (colourBitMultiplier > (width * height)) {
          var diff = (colourBitMultiplier / height) - width;

          var f = 0;
          while (f < decoded.length) {
              newFix += decoded.substr(f, width);
              f += width + diff;

          }

          decoded = newFix;
      }

      //console.log("about to decode");
      var imageData;
      if (colorbits === 8) {
          // imageData = show.decodeBMP256(decoded, show.tempShowContext);
          imageData = show.decodeBMP256(decoded);
      } else {
          //Changed this to include upside down flag
          // imageData = show.decodeBMP(decoded, show.tempShowContext, show.isUpsideDown);
          imageData = show.decodeBMP(decoded, show.isUpsideDown);
      }
      //console.log("about to get x and y");
      xpos = this.FF(6+0) - show.xOffset;
      ypos = this.FF(6 + 2) - show.yOffset;

      show.PushFragment(xpos, ypos, imageData);
  };

  nc.StudentUI = function(){
    //nc_studentUi
    var d = this.d;
    var cmdSet = d.charCodeAt(14); // cmdSet @ 6 - Maybe?

    if (cmdSet === 1 || showFeedBack === true) {
        this.StudentFeedback(d);
        // nc.StudentFeedback(d);
    } else if (cmdSet === 2){
        // this.QandA();
        nc.QA(d);
    } else {
      // Do nothing
    }
  };

  nc.QA = function(d){
      //nc_studentUi
      QA = this.csParams.QA;
      QA.Cid = this.cid;

      var cmdSet = d.charCodeAt(6); // New
      if (cmdSet === 2) {
        if (d.charCodeAt(22) == 2) {
            Tab.Make('qanda.html', 'QandA');
          }

        var includesMoreData = d.charCodeAt(24);
        if (includesMoreData && d.length > 25) {
            QA.AnswerBtnAvailable = d.charCodeAt(28);
            QA.ThinkingTime = d.charCodeAt(32);
            QA.TimeLimit = d.charCodeAt(36);
            QA.PlaySounds = d.charCodeAt(38);
            QA.RoundNum = d.charCodeAt(41);
        }

        this.csParams.QA = QA;
      }
  };

  nc.StudentFeedback = function(d) {
    //nc_studentUi
    Feedback = this.csParams.Feedback;
    Feedback.Cid = this.cid;

    if (showFeedBack === false) {
      showFeedBack = true;
    } else {
      if (d.charCodeAt(17) === 1) {
        if (d.charCodeAt(24) == 1) {
          Tab.Make('feedback.html', 'Feedback');
        } else {
          showFeedBack = false;
          Tab.Remove("Feedback");
        }
      }
    }

    if (d.length > 28) {
      Feedback.ImageSet = d.charCodeAt(28);
      Feedback.NumberAnswers = d.charCodeAt(32);
      Feedback.Question = d.substring(36).split(String.fromCharCode(0))[0];
      Feedback.AnswerText = d.substring(36).split(String.fromCharCode(0))[1];
      Feedback.ButtonConfiguration = d.charCodeAt(d.length-8);
    }

    this.csParams.Feedback = Feedback;
  };

  nc.Message = function(){
      debug('in MESSAGE');

      var d = this.d;
      var i = {
        sender: d.indexOf(z, 4 + 2),
        msg: d.indexOf(z, 4 + 18),
        delimiter: undefined,
        terminator: undefined
      }

      i.delimiter = d.indexOf('|', i.msg + 1);
      i.terminator = d.indexOf(z, i.delimiter + 1);
      // var termIndex1 = d.indexOf(z, 4 + 2); // End of header
      // var termIndex2 = d.indexOf(z, 4 + 18); // End of sender
      // var termIndex3 = d.indexOf('|', termIndex2 + 1); // Message Delimiter
      // var termIndex4 = d.indexOf(z, termIndex3 + 1); // message Terminator
      var message;
      sender = d.substring(4 + 2, i.sender);

      try {
        sender = UTF8.Decode(d.substring(4+2, i.sender));          
      } catch(error) {
        console.log('Failed to decode',d.substring(4+2, i.sender),error);
        sender = "\uac15\uc0ac";
      }

      zx = i.terminator + 2;
      zxn = d.indexOf(z, zx);
      if (zxn >=0)
      {
          sender = UTF8.Decode(d.substring(zx, zxn), this.cid);
      }

      message = UTF8.Decode(d.substring(4 + 18, i.msg), this.cid);

      // intercept plaintext message
      // (ncJson will return true if it can delegate)
      // if(nc.ncJson(message)) return;

      var timeout = d.substring(i.msg + 1, i.delimiter);

      var title = TRANSLATE('MESSAGE_FROM_PLACEHOLDER', sender);

      var dt = new Date();
      var dtstr = dt.getDate()+'/'+(dt.getMonth()+1)+'/'+dt.getFullYear();
      var tmstr = dt.toLocaleTimeString();
      var prefix = '(' + dtstr + ', ' + tmstr + '): ';

      message = prefix + message;
      var messages = [];
      var msgStart = 0;
      var msgSpace = (title.length > 28)? 116 : 116; // Chrome Notification Limits;

      while (msgStart>-1)
      {
          if (msgStart + msgSpace > message.length)
          {
              messages.push(message.substring(msgStart));
              msgEnd = -1;
          }
          else
          {
              msgEnd = message.substring(msgStart, msgStart+msgSpace).lastIndexOf(' ');
          }

          if (msgEnd !== -1)
          {
              messages.push(message.substring(msgStart, msgStart+msgEnd));
              msgStart += ++msgEnd;
          }
          else {
              msgStart = msgEnd;
          }
      }

      var msg = messages.pop();
      while (msg !== undefined)
      {
          MessageDialog(title, msg, timeout);
          msg = messages.pop();
      }

      /*
      var notification = webkitNotifications.createNotification('images/message48.png', title, '(' + dtstr + ', ' + tmstr + '): ' + message);
      notification.show();
      if (timeout > 0) {
          window.setTimeout(function() { notification.cancel(); }, timeout * 1000);
      }

      */
  };

  function MessageDialog(title, message, timeout)
  {
      var opt = {
          body: message,
          icon: 'images/message48.png'
      };

      var notifi = new Notification(title, opt);
      if(timeout){
          window.setTimeout.call(window,() => notifi.close(window), timeout * 1000);
      }
  }

  nc.Plugin = function(){
      // nc_plugin
      var d = this.d;
      var tabId;
      // console.log("plugin",d.charCodeAt(4+2),`cmd:${d.charCodeAt(4 + 10)}, pcmd:${d.charCodeAt(4 + 14)}`);
      if (d.charCodeAt(4 + 2) == 2) // webid
      {
          cmd = d.charCodeAt(4 + 10);
          pcmd = d.charCodeAt(4 + 14);
          if (cmd == 3) // init
          {
              //alert("webinit");
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 5) // init filter
          {
              zx = d.indexOf(z, 4 + 38);
              var webfilter = d.substring(4 + 38, zx);
              //alert("webfilter=" + webfilter);
               // alert("web poll");
              this.csParams.Connections[this.cid].webpoll = 1;
              Tab.parseWebFilter(this.cid, webfilter);
              Tab.filterTabs();
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 6) // poll
          {
              // alert("web poll");
              this.csParams.Connections[this.cid].webpoll = 1;
              // Trigger polling mechanism

              Tab.PollActive();
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 2) // close
          {
              tabId = d.charCodeAt(4 + 38) + (d.charCodeAt(4 + 39) << 8) + (d.charCodeAt(4 + 40) << 16) + (d.charCodeAt(4 + 41) << 24);
              //alert("web close, tabid=" + tabid);
              //if we support multiple tabs we can just remove the tabid
              chrome.tabs.remove(tabId);

              // but since we only support a single tab, we need to remove the currently selected one
              // removeSelectedTab();
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 1) // start (open url)
          {
              termIndex = d.indexOf(String.fromCharCode(0), 4 + 38);
              var url = d.substring(4 + 38, termIndex);
              //alert("web open " + url);
              Tab.MakeExternal(url);
          }
          else {
            debug("web plugin, cmd=" + cmd + ", pcmd=" + pcmd);
          }
      }
      else if (d.charCodeAt(4 + 2) == 1) // appid
      {
        var tabid;
          cmd = d.charCodeAt(4 + 10);
          pcmd = d.charCodeAt(4 + 14);
          if (cmd == 3) // init
          {
              //alert("appinit");
              var app = String.fromCharCode(protocol.PLUGIN) + z + String.fromCharCode(1) + z3 + z4 + String.fromCharCode(3) + z3 + z4 + z4 + z4 + z2 + z;
              extension.senddata(this.cid, app);

              var web = String.fromCharCode(protocol.PLUGIN) + z + String.fromCharCode(2) + z3 + z4 + String.fromCharCode(3) + z3 + z4 + z4 + z4 + z2 + z;
              extension.senddata(this.cid, web);

              var web2 = String.fromCharCode(protocol.PLUGIN) + z + String.fromCharCode(2) + z3 + String.fromCharCode(3) + z3 + String.fromCharCode(7) + z3 + z4 + String.fromCharCode(5) + z3 + z4 + String.fromCharCode(5) + z + z;
              extension.senddata(this.cid, web2);
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 6) // poll
          {
              //alert("app poll");
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 3) // activate
          {
            tabId = d.charCodeAt(4 + 40) + (d.charCodeAt(4 + 41) << 8) + (d.charCodeAt(4 + 42) << 16) + (d.charCodeAt(4 + 43) << 24);

            chrome.tabs.update(tabId, {active: true, highlighted: true }, function(tab){
              chrome.windows.update(tab.windowId, {focused: true, drawAttention: true});
            });
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 5) // init filter
          {
          }
          else if ((cmd == 7 || cmd == 13) && pcmd == 2) // close app
          {
            tabId = d.charCodeAt(4 + 40) + (d.charCodeAt(4 + 41) << 8) + (d.charCodeAt(4 + 42) << 16) + (d.charCodeAt(4 + 43) << 24);
            //alert("web close, tabid=" + tabid);
            //if we support multiple tabs we can just remove the tabid
            chrome.tabs.remove(tabId);
          }
          else {
              debug("app plugin, cmd=" + cmd + ", pcmd=" + pcmd);
          }
      }
  };

  nc.Chat = function(){
      debug('in CHAT');

      var d = this.d;
      var msg = '';
      var chat = Chat.instance();

      var array = this.d.split("").map(c => c.charCodeAt(0))
    //   console.log("CHAT", array)
      let i = {sender:6+8, msg:6+24}
      var zx = d.indexOf(z, i.sender);
      var zx2 = d.indexOf(z, i.msg);
      var zx3 = d.indexOf(z, zx2 + 1);

    //   console.log(contents(d));
    //   console.log("Indices:",zx,zx2,zx3);

      //flags = d.charCodeAt(6 + 4) + (d.charCodeAt(6 + 5) << 8) + (d.charCodeAt(6 + 7) << 24);
      flags = this.FF(6+4)+(d.charCodeAt(6+7)<<24);
      var submode = (d.charCodeAt(6 + 6) << 16);
      // chat.chatID= d.substring(6, 10);
      // chat.chatCID = this.cid;

      // debug("Flags: ",flags);
      // debug("subMode ", submode);#'#'
      sender = d.substring(i.sender, Math.min(zx,i.msg));
    //   console.log(contents(d));
    //   console.log(contents(d.substring(6 + 8, Math.min(zx,i.msg))));

      if (zx <= i.msg) {
        try {
          sender = UTF8.Decode(d.substring(i.sender, Math.min(zx,i.msg)));          
        } catch(error) {
          console.log('Failed to decode',d.substring(i.sender, Math.min(zx,i.msg)),error);
          sender = "\uac15\uc0ac";
        }        
      } else if (zx > i.msg){
        sender = UTF8.Decode(d.substring(i.sender, zx))
      }
      
      if (zx2 !== -1)
      {
          msg = UTF8.Decode(d.substring(6 + 24, zx2),this.cid);
      }
      if (zx3 !== -1 && zx3 > zx2 + 1)
      {
          sender = UTF8.Decode(d.substring(zx2+1, zx3), this.cid);
      }

      switch (flags) {

          case 1: // start chat (from sender), chatid is unique id (needed when we send chat data back to Tutor)
              chat = new Chat(d.substring(6, 10), this.cid);
              Chat.instance(chat);
              chat.setTitle("");
              if (submode==65536) { // topic is msg
                chat.setTitle(msg);
              }
              if (submode & 2) { // chat is optional (client can Join or Decline)
              }
              if (submode & 4) { // chat is always on top
              }
              if (submode & 8) { // chat can't be closed
              }
              chat.startChat();
              chat.sendChatJoin();
              //console.log(msg);
              break;
          case 2: // sender has joined
              chat.addJoinedMessage(sender);
              break;
          case 3: // sender has left
              chat.addLeftMessage(sender);
              break;
          case 4: // sender says msg
              chat.addMessage(sender, msg);
              break;
          case 5: // close chat
              if (chat.isChatting)
              {
                  // Only display if we didn't request the closure.
                  MessageDialog('', TRANSLATE('TUTOR_HAS_LEFT_CHAT'), 8);
              }
              chat.eject();
              break;
          case 13: // history: sender says
              tmpSaysLowerCase = TRANSLATE('says').charAt(0).toLowerCase() + TRANSLATE('says').slice(1)
              if (msg.indexOf(tmpSaysLowerCase) > -1) {
                  tmpSender = msg.substring(0, msg.indexOf(tmpSaysLowerCase));
              } else {
                  if (tmpSender!=="" && msg !=="" && msg !==undefined){
                      chat.addMessage(tmpSender, msg);
                      tmpSender="";
                  }
              }
          break;
      }

      /*TODO: MAKE THIS WORK
      zx = d.indexOf(z, 6 + 8);
      sender = d.substring(6 + 8, zx);
      var zx2 = d.indexOf(z, 6 + 24);
      var msg = d.substring(6 + 24, zx2);
      flags = d.charCodeAt(6 + 4) + (d.charCodeAt(6 + 5) << 8) + (d.charCodeAt(6 + 7) << 24);
      var submode = (d.charCodeAt(6 + 6) << 16);
      if (msg!==""){
          chat.addMessage(sender, msg);
          extension.senddata(cid, d);
      }

      //wip alert("chat, sender=" + sender + ", msg=" + msg);

      //wip chrome.browserAction.setIcon({path:"chat.gif"});
      //wip chrome.browserAction.setPopup({popup:"chat.html"});*/
  };

  nc.ApprovedWebList = function() {
      // nc_approvedweblist
      flags = this.d.charCodeAt(6);
      x = 6 + 16;
      zx = this.d.indexOf(z, x);
      var list = this.d.substring(x, zx);
      if (flags & 1)
          gApproved = '';
      // if (flags & 2) {

      gApproved = gApproved + list.replace(',', '.');
      localStorage.approvedWebList = gApproved.split('\n');
      //}
      //alert("approved web list=\n" + list);
  };

  nc.WinMouse = function() {
      // nc_winmouse
      var nevents = this.FF(6 + 0);
      var idx = 6 + 2;
      while (nevents--)
      {
          xpos = this.d.charCodeAt(idx) + (this.d.charCodeAt(idx + 1) << 8);
          ypos = this.d.charCodeAt(idx + 2) + (this.d.charCodeAt(idx + 3) << 8);

          if (xpos == 32767) { // click using ypos as buttonstate
              Mouse.click(ypos);
          }
          else if (xpos == 32766) { // wheel mouse
              if (ypos >= 32768) {
                  ypos = ypos - 65536;
              }
              chrome.tabs.executeScript(null, {code: 'window.scrollBy(0,' + -ypos + ');'});
          }
          else { // mouse move
              Mouse.x = xpos;
              Mouse.y = ypos;
          }
          idx += 4;
      }
  };

  nc.Thumbnail = function() {
      // nc_thumbnail
      connection = CSParams.Connections[this.cid];

      if (!connection) {
        console.warn("Specified cid not in conntections - have we reset?", this.cid, connnection);
        // Specified cid not in conntections - have we reset?
      }
      flags = this.d.charCodeAt(6) + (this.d.charCodeAt(7) << 8);
      if (flags & 1){
        connection.thumbnailType = "image/jpeg";
      }

      var thumbi;
      thumbi = this.d.charCodeAt(4 + 11) * 256 + this.d.charCodeAt(4 + 10); // interval
      connection.thumbi = thumbi;
      if (thumbi === 0){
        //   console.log("clearing intervalThumbnailTimer");
          // Clear interval prior to changing dimensions
          window.clearInterval(thumbnailTimer);
      }
      connection.thumbs = thumbi; // since last shot, so that takeScreenshot is immediate
    // console.log(connection.thumbw, connection.thumbh);
      connection.thumbw = this.d.charCodeAt(4 + 7) * 256 + this.d.charCodeAt(4 + 6); // width
      connection.thumbh = this.d.charCodeAt(4 + 9) * 256 + this.d.charCodeAt(4 + 8); // height
    //   console.trace();
    //   console.debug('Thumbnail dimension update:',connection.thumbw, "height:", connection.thumbh);
      connection.thumblast = ''; // last
      if (thumbi === 0){
          // Already Handled
      } else {
          //console.log("setting Timer");
          window.clearInterval(thumbnailTimer);
          Stream.enable();
          Stream.confirmStream()
          .then(function(payload) {
            // console.log('nc.Thumbnail SendStream',connection.thumbw, connection.thumbh);
            nc.SendWSlave(payload);
          });

          thumbnailTimer = setInterval(() => {
            Stream.confirmStream()
            Image.takeScreenshot()
              .then(extension.sendImageData)
              .catch(err => console.log('invalid thumbnail size', err));
          },thumbi * 1000);
      }

      // if (tcpConnected){
      //     Image.takeScreenshot();
      //      chrome.windows.getCurrent(function(w) {
      //         gThumbWidth = this.d.charCodeAt(4 + 7) * 256 + this.d.charCodeAt(4 + 6); // width
      //         gThumbHeight = parseInt(gThumbWidth / (w.width/w.height), 10);
      //
      //         if (thumbi === 0){
      //             window.clearTimeout(thumbnailTimer);
      //         }else{
      //             thumbnailTimer = window.setInterval(Image.takeScreenshot,thumbi*1000);
      //         }
      //     });
      // }

      if (thumbi) {
          Subscription.on.interval.call();
      }
  };

  nc.ChromeWindow = function(view, w, sizeModifier, dataBuilder)
  {
      view.width = w.width - 20 + sizeModifier;
      view.height = w.height - 101 + sizeModifier;

      view.gridSizeX = view.width;
      view.gridSizeY = 32;

      view.canvas.width = view.gridSizeX;
      view.canvas.height = view.gridSizeY;

      view.nrX = Math.round(view.width / view.gridSizeX);
      view.nrY = Math.round(view.height / view.gridSizeY);

      view.pixelCount = (view.gridSizeX * view.gridSizeY) * 4;
      width = view.width;
      height = view.height;

      data = dataBuilder(width, height);
      extension.senddata(this.cid, data);
  };

  nc.SendTabData = function()
  {
    let connection = CSParams.Connections[this.cid];
    Tab.getAllTabsFromOpenWindows()
    .then(function (tabs) {
      connection.force = true;
      tabs.forEach(Tab.SendData);
      connection.force = false;
    });

    CSParams.get.currentWindow()
    .then(Tab.updateCurrentWebsite)
  };

  nc.SendWSlave = function(w)
  {
    let { width, height } = w;
    // wslave`
    console.debug('PROTOCOL.SendWSlave:',width, height);
    let dataBuilder = (width, height) => String.fromCharCode(protocol.WSLAVE) + z + z2 + String.fromCharCode(16) + z + z2 + String.fromCharCode(width & 255) + String.fromCharCode((width >> 8) & 255) + String.fromCharCode(height & 255) + String.fromCharCode((height >> 8) & 255) + z2;
      return new Promise((resolve,reject) => {
          // send WSLAVE once to tell Control how big the View window should be
          // Why 4?
          nc.ChromeWindow(View, w, 4, dataBuilder);
          resolve(View);
      });
  };

  // LEGACY
  // nc.SendWSlave = function(w)
  // {
  //   // wslave`
  //   let dataBuilder = (width, height) => String.fromCharCode(protocol.WSLAVE) + z + z2 + String.fromCharCode(16) + z + z2 + String.fromCharCode(width & 255) + String.fromCharCode((width >> 8) & 255) + String.fromCharCode(height & 255) + String.fromCharCode((height >> 8) & 255) + z2;
  //     return new Promise((resolve,reject) => {
  //         // send WSLAVE once to tell Control how big the View window should be
  //         nc.ChromeWindow(View, w, 4, dataBuilder);
  //         resolve(View);
  //     });
  // };

  function getAllTabsFromOpenWindows()
  {
    return new Promise(
      function(resolve) {
      chrome.windows.getAll({ populate: true}, function(w) {
        resolve(w
          .map(browserWindow => browserWindow.tabs)
          .reduce((acc,currentValue) => acc.concat(currentValue)));
      });
    });
  }

  return nc;
}));
