(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory();
  } else {
      // Browser globals (root is window)
      root.NSLibrary = root.NSLibrary || { Comms: { NC:{ }}};
      root.NSLibrary.Comms.NC.protocol = factory();
  }
}(this, function() {
  return {
        NAMEREQ:         1,            /* request slave name            */
        NAMERESP:        2,            /* slave name is...              */
        SELECT:          3,            /* select slave                  */
        ACCEPT:          4,            /* accept call                   */
        ACCEPTSEL:       5,            /* accept select                 */
        REJECT:          6,            /* reject call/select            */
        SETMODE:         7,            /* set slave mode                */
        REFRESH:         8,            /* resend slave screen           */
        KEYSTROKES:      9,            /* keystroke(s)                  */
        SCREENDATA:     10,            /* screen update data            */
        BEEP:           11,            /* beep                          */
        MESSAGE:        12,            /* message                       */
        CHATKEYS:       13,            /* chat keystrokes               */
        KILLSLAVE:      14,            /* de-activate slave             */
        CHATREQ:        15,            /* slave wants to chat           */
        WCHATREQ:       16,            /* Windows slave wants to chat   */
        UNAVAIL:        17,            /* Slave in graphics mode        */
        KILLCHAT:       18,            /* dismiss chat                  */
        PALETTE:        19,            /* EGA palette registers         */
        DAC:            20,            /* VGA DAC registers             */
        POPDOWN:        21,            /* remove Windows message/chat   */
        FULLSCREEN:     21,            /* WIN32 app running full screen */
        MOUSE:          22,            /* mouse state change            */
        BIOSKEYS:       23,            /* BIOS keystroke(s)             */
        LISTDRIVES:     24,            /* list drive letters            */
        DRIVES:         25,            /* drive letters response        */
        LISTFILES:      26,            /* list files/directories        */
        FILES:          27,            /* files/directories response    */
        GETDIR:         28,            /* get current drive and dir     */
        CURRDIR:        29,            /* current drive & dir response  */
        OPENFILE:       30,            /* open file                     */
        CLOSEFILE:      31,            /* close file request            */
        READFILE:       32,            /* read file request             */
        FILEDATA:       33,            /* read file response            */
        WRITEFILE:      34,            /* write file request            */
        DELETEFILE:     35,            /* delete file request           */
        SETFATTR:       36,            /* set file attributes           */
        MKDIR:          37,            /* create directory              */
        RMDIR:          38,            /* delete directory              */
        RENAME:         39,            /* rename file                   */
        SEEKFILE:       40,            /* seek in file                  */
        FILERESULT:     41,            /* file I/O response             */
        BSMESSAGE:      42,            /* message in broadcast show     */
        PRINT:          43,            /* print characters supplied     */
        PRINTACK:       44,            /* characters have been printed  */
        REBOOT:         45,            /* reboot slave                  */
        WINMOUSE:       46,            /* Windows mouse state change    */
        WINDATA:        47,            /* Windows screen data           */
        POPUP:          48,            /* Windows slave popup menu      */
        WSLAVE:         49,            /* setmode response from WSLAVE  */
        FLUSH:          50,            /* Flush slave cache             */
        KBDLIGHTS:      51,            /* Keyboard lights (for Windows) */
        SLISTFILES:     52,            /* list files/directories (short)*/
        SFILES:         53,            /* files/directories response    */
        LLISTFILES:     54,            /* list files/directories (long) */
        LFILES:         55,            /* files/directories response    */
        LGETDIR:        56,            /* get current drive and dir     */
        LOPENFILE:      57,            /* open file                     */
        LCLOSEFILE:     58,            /* close file request            */
        LDELETEFILE:    59,            /* delete file request           */
        LSETFATTR:      60,            /* set file attributes           */
        LMKDIR:         61,            /* create directory              */
        LRMDIR:         62,            /* delete directory              */
        LRENAME:        63,            /* rename file                   */
        PRINTJOB:       64,            /* print job - V2.70 and later   */
        ECHO_IN:        65,            /* echo data - V2.70 and later   */
        ECHO_OUT:       66,
        CAD:            67,            /* ctrl-alt-del, for NT logon    */
        SDATA:          68,            /* packet stream                 */
        EDATA:          69,            /* end of uncompressed stream    */
        ECDATA:         70,            /* end of compressed stream      */
        DCAFDATA:       71,            /* From the OS/2 client          */
        MMCMD:          72,            /* MMPlayer stuff                */
        MMDATA:         73,
        SOS:            74,            /* Client wants help             */
        SOS_CLEAR:      75,            /* No she doesn't                */
        PRDRIVER:       76,            /* Printer driver request        */
        PD_ACK:         77,            /* Reply to same                 */
        SYSINFO:        78,            /* Get system info req/reply     */
        EXEC:           79,            /* Run a program                 */
        EXECRESULT:     80,            /* Did it start?                 */
        LOGIN:          81,            /* Send user ID and password     */
        CONFIG:         82,            /* Returned if login succeeds    */
        REGREQUEST:     83,            /* see NC_REGxxx #defines below  */
        REGRESULT:      84,            /* Result from OpenKey and such  */
        REGKEYS:        85,            /* Registry keys returned        */
        REGVALS:        86,            /* Registry values returned      */
        REGVALOVERFLOW: 87,            /* If value too long             */
        ALIASES:        88,            /* Aliases for FTRAN, esp Desktop and Documents */
        CALL:           90,            /* call slave                    */
        HANGUP:         91,            /* hangup session                */
        ACK:            92,            /* acknowledge packet            */
        NACK:           93,            /* ask for packet to be re-sent  */
        CRC32:          99,            /* tcp data is checksummed with Crc32, esp for wireless */
        GETMOUSE:      100,            /* for WSLAVE                    */
        WTOA:          101,            /* for CLHOOK4                   */
        KILLPIPES:     102,            /* from GDIHOOK5                 */
        HOOKDATA:      103,            /* from CLHOOK4, tells client32 to send some data to gdi */
        CTLQUERY:      110,            /* query Control status          */
        CTLSTATUS:     111,            /* return Control status         */
        CONNECTREQ:    112,            /* request a connection          */
        AUDIOFMT:      113,            /* send audio format to Slave    */
        AUDIOERR:      114,            /* Slave can't handle audio fmt  */
        AUDIO:         115,            /* audio in agreed format        */
        USERLOGIN:     116,            /* Displays box at client prompting for user name */
        CANCELLOGIN:   117,            /* Cancel the user login box */
        DESPKT:        118,            /* DES encrypted data */
        CLIPDATA:      119,            /* Clipboard data */
        GETCHECKSUM:   120,            /* Get file checksum */
        CHECKSUM:      121,            /* file checksum returned */
        INVALRECT:     122,            /* invalidate rect */
        DCAFPAL:       123,            /* DCAF palette */
        DCAFINFO:      124,            /* DCAF version of NC_WSLAVE */
        INCLUDERECT:   125,            /* Include rectangle for Show (clips to this) */
        VNCDATA:       126,            /* A packet of VNC data */
        TICKLE:        127,            /* is the other end still alive? */
        VNCCLIP:       129,            /* VNC clipboard data */
        CHATEX:        130,            /* Group Chat */
        LOADCTL:       131,            /* Tell the client to load the control */
        SHOWLEADER:    132,            /* for client to take lead of control that's showing to it */
        ANNOTATE:      133,            /* Start or stop annotating at client while viewing */
        COMPONENT:     134,            /* Interact with a component at the client */
        ENCRYPTED:     135,
        PROCESSINFO:   136,
        PUBLISHCLASS:  137,
        CLOSEAPP:      138,            /* Close the specified app */
        UDP_CONTROL:   139,            /* PaulS, 13Feb02, for Broadcast Show */
        TASKLIST:      140,            /* Request client to return list of processes */
        PROCESSCTRL:   141,            /* Request client to kill process */
        PROCESSLIST:   142,
        READCHECKSUMS: 143,            /* Read checksums from file */
        CHECKSUMDATA:  144,            /* the checksum data */
        COPYFILEDATA:  145,            /* copy data from one file to another on the same client */
        INVENTORY:     146,
        READFILEEX:    147,            /* read file with seek */
        VOTE:          148,            /* Vote at client*/
        SHOWVIDEO:     149,            /* Show video at client */
        PLUGIN:        150,            /* Interact with a plugin at the client */
        PLAYERCMD:     151,            /* NSS Player Commands TPD 16/10/02 */
        PRESHOWVIDEOCHECK: 152,        /* Pre Show video check at client Ditta 20/02/03*/
        DRIVESHIELD:   153,
        BIGNEASY:      154,            /* used by connect.cpp to magically split up large packets and concatenate them back together */
        THUMBNAIL:     155,            /* Send info for thumbnails & receive thumbnail data */
        APPROVEDWEBLIST: 156,          /* Send  Approved web list data to Clients  */
        SESSION:       157,           /* session id - so students can reconnect back to a class/session */
        RECONNECT:     157,           /* session id - so students can reconnect back to a class/session */
        NEWSCRAPE:     158,           /* new scrape result, for nsm 9. See also NC_NEWSCRAPEEX */
        WHITEBOARD:    159,
        REMOTECMD:     160,           /* remote command prompt, see h\nc_remcmd.h */
        WB_LEADER:     161,
        REGISTER_USER:   162,          /*User Registeration data */
        CANCEL_REGISTER: 163,
        UNREGISTER_USER: 164,
        USERCHANGED:   165,
        AUTOLOGON:     166,           /* Auto client logon */
        COBROWSE:      167,           /* Cobrowse; see NSMWBComms */
        COBROWSECMD:   168,           /* Cobrowse; see NSMWBComms */
        SETCLIP:       169,           /* Set Clip Region - for nss850 */
        NSM_PROTECT:   170,           /* Control NetSupport protect (or an oem) */
        THUMBNAILEX:   171,           /* enhanced to return monitor count and some sys info bits */
        GETMONITORS:   172,           /* returns monitor details at client */
        STUDENTUI:     173,
        MESSAGEEX:     174,
        MUTESOUND:     175,
        TYPING:        176,
        SMARTCARD:     177,
        RDP:           178,
        SETMON:        179,
        VOTESHOWRESULT: 179,
        JOURNAL:       180,
        GIVE_REWARD:   181,
        LINKTUTOR:     182,
        BATTERY:       183,
        WIRELESS:      184,
        ANNOUNCE:      185,
        AUDIOSTAT:     186,
        SETVOLUME:     187,
        CHANNELUPDATE: 188,
        VIDEOHINT:     189,
        NEWLIC:        190,
        THUMBNAILFILE: 191,
        UDPFILESTART:  192,
        UDPFILEDATA:   193,
        UDPFILEEND:    194,
        LATLONG:       215,
        NS_JSON:       235,         /* JSON String (eg. [CID, 4 bytes][NC_JSON, 2 bytes][unused, 2 bytes][PAYLOAD])*/
    };
}));
