/*
Hard dependancies: (external)
////////////////////////////////////////////////////////
gUsedIPAddress              chromeguts.js
NSLibrary.UTF8              utf8.js
NSLibrary.Comms             comms.js
str2ab                      comms.js
localStorage                window
btoa                        window


PROPERTIES of signatures:
////////////////////////////////////////////////////////
'ACK',
'APPTYPE',
'AT',
'CHANNEL',
'CLIENT_ADDR',
'CLIENT_NAME',
'CLIENT_VERSION',
'CMD',
'CONNECTION_ID',
'DATA',
'FLAGS',
'FROM',
'GSK',
'HOSTNAME',
'LEN',
'MACADDRESS',
'PROTOCOL_VER',
'REQUESTING_HELP',
'RESULT',
'TS',
'USERNAME'
*/


(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['comms','utf8', 'modules/window'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('comms'),
        require('utf8'),
        root
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
      // var NSLibrary = NSLibrary || {};
      root.NSLibrary.Comms = NSLibrary.Comms || {};
      root.NSLibrary.Comms.Websocket = factory(
        NSLibrary.Comms,
        NSLibrary.UTF8,
        root);
  }
}(this, function(Comms, UTF8, self) {

  const { localStorage, btoa } = self;
  const states = ["CONNECTING", "OPEN", "CLOSING", "CLOSED"];

  class WebsocketCommands extends Comms {
      constructor(mechanism, websocket){
        super(mechanism);
        wsComms.mechanism = mechanism;
        wsComms.instance = this;

        this.websocket = websocket;
      }

      OPEN() {
        const { macAddress, gatewayKey } = localStorage;
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "OPEN",
            CLIENT_VERSION: "1.0",
            PROTOCOL_VER: "1.1",
            CLIENT_NAME: WebsocketCommands.clientName(),
            CLIENT_ADDR: ">255.255.255.255",
            HOSTNAME: WebsocketCommands.clientName(),
            MACADDRESS: WebsocketCommands.prefix(macAddress),
            GSK: WebsocketCommands.prefix(gatewayKey),
            APPTYPE: 4096
          })
        );
      }

      HEARTBEAT() {
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "HEARTBEAT",
            APPTYPE: 4096,
            REQUESTING_HELP: 0
          })
        );
      }

      POLL(ts) {
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "POLL",
            ACK: 1,
            TS: ts
          })
        );
      }

      STATUS(room, requestingHelp) {
        return this._sendAsArrayBuffer(
          WebsocketCommands.serialize({
            CMD: "STATUS",
            APPTYPE: 4096,
            REQUESTING_HELP: requestingHelp,
            USERNAME: WebsocketCommands.clientName(),
            CHANNEL: WebsocketCommands.prefix(room)
          })
        );
      }

      HANGUP(cidt) {
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "HANGUP",
            CONNECTION_ID: cidt
          })
        );
      }

      CONNECT_REPLY(cidt) {
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "CONNECT_REPLY",
            CONNECTION_ID: cidt,
            CLIENT_NAME: WebsocketCommands.clientName(),
            RESULT: 0
          })
        );
      }

      BROADCASTDATA(room, nc_announce) {
        // if (!this._isReadyState(1)) return;
        // TODO:: COMPLETE THIS IMPLEMENTATION
        return this._sendBase64Ascii(
          WebsocketCommands.serialize({
            CMD: "BROADCASTDATA",
            FLAGS: 0,
            AT: 2560,
            FROM: "NCS/>" + gUsedIPAddress,
            CHANNEL: WebsocketCommands.prefix(room),
            LEN: nc_announce.length,
            DATA: yEncode(nc_announce)
          })
        );
      }

      CLOSE() {
        return this._sendBase64Ascii(WebsocketCommands.serialize({ cmd: "CLOSE" }));
      }

      static prefix(string) {
        return UTF8.PrefixEncode(string);
      }

      static clientName() {
        return this.prefix(localStorage.clientName);
      }

      static serialize(object) {
        // return transmission string
        // Object.key MUST BE uppercase
        let array = Object.keys(object).map(key => `${key.toUpperCase()}=${object[key]}`);

        array.push(""); // terminate linebreak
        return array.join("\n");
      }

      static convertBinaryToString(message, length)
      {
          let d = '';

          for (var i = 0; i < length; ++i)
          {
              d += String.fromCharCode(message[i]);
          }

          return d;
      }

      /**
       * PROTECTED METHODS
       */

      _sendBase64Ascii(string) {
        if (!this._isReadyState(1)) return;
        const ascii = btoa(string);
        this.websocket.send(ascii);
        return ascii;
      }

      _sendAsArrayBuffer(string) {
        if (!this._isReadyState(1)) return;
        const array = Comms.str2ab(string);
        this.websocket.send(array);
        return array;
      }

      _isReadyState(n) {
        // Retrun boolean
        // Check that the n state matched the current state
        // a mismatch will print the current state.
        const {readyState} = this.websocket;
        const match  = readyState === n;
        const msg = `Socket is (${states[readyState]}) not (${states[n]})`;
        if (!match) {
          console.warn(msg);
        }
        return match;
      }
  }

  function yEncode(data)
  {

      var res = [];
      var r=0;

      for (var i = 0; i < data.length; ++i)
      {
          switch(data.charCodeAt(i) +42) {
              case 0:
              case 9:
              case 10:
              case 13:
              case 256:
              case '=':
              case '.':
                  res[r++] = '=';
                  res[r++] = String.fromCharCode((data.charCodeAt(i) + 64 + 42) % 256);
                  break;
              default:
                  res[r++] = String.fromCharCode((data.charCodeAt(i) + 42) % 256) ;
                  break;
          }
      }

      return res.join("");
  }

  /**
   * instance - Currently active instance.
   * mechanism - Allows you to control and manipulate the currently active socket mechanism
   * Commands - Class, holds static methods.
   * send - transmit using the current instance.
   */

  var wsComms = {
    instance: undefined,
    mechanism: undefined,
    Commands: WebsocketCommands,
    send
  };

  function send(data)
  {
    wsComms.instance.websocket.send(data);
  }

  // const { yEncode, localStorage, btoa, str2ab, gUsedIPAddress } = window;
  // const { send, readyState } = websocket;
  // const { UTF8.PrefixEncode } = UTF8;
  // wsComms.Commands = WebsocketCommands;

  return wsComms;
}));

/**
 * METHOD NOT CALLED ANYWHERE
 */

// function sendPinCommand(pin)
// {
//     var toSend = "CMD=CLIENT_PIN_REQUEST\nDATA=LOC_0/HTTP>"+ gUsedIPAddress +";MAC_0/" + gUsedMacAddress+ ";\nCLIENT_NAME="+UTF8.PrefixEncode(localStorage.clientName)+"\n";
//     if (pin){
//         toSend += "PIN="+pin+"\n";
//     }
//     if (localStorage.pinServerAddress && localStorage.pinServerPort){
//         pinWebSocket = new WebSocket("ws://" + localStorage.pinServerAddress + ":" + localStorage.pinServerPort);
//
//         pinWebSocket.onopen = function() {
//             pinWebSocket.send(btoa('CMD=OPEN\nCONSOLE=1\nPROTOCOL_VER=1.1\nAPPTYPE=4096' + '\n'));
//             window.setTimeout(    function(){   pinWebSocket.send(btoa(toSend));}    , 500);
//
//         };
//
//         pinWebSocket.onmessage = function(evt) {
//             var d = atob(evt.data);
//             var parts;
//             if(d.substr(0,14)=="CMD=PIN_RESULT"){
//                 parts = d.split('\n');
//                 gPin = parts[1].replace("PIN=","");
//                 pinWebSocket.close(1000);
//
//
//             }else if(d.substr(0,20)=="CMD=PIN_CLEAR_RESULT"){
//                 parts = d.split('\n');
//                 var result = parts[1].replace("RESULT=","");
//                 debug(result);
//             }
//         };
//     } else {
//         WSComms.instance.websocket.send(btoa(toSend));
//     }
// }
