/*
    Console log index by color

    Creates a window method that returns an Array of
    arguments to colour code the console.log output
    Each log has a timestamp, and time diff of the last log,
    the log message is uniquely coloured according
    to the 1st string in the argument list

    Usage:
    console.log.apply(console, _c("Log identifier", optional, values));

*/

(function(root, factory) {
    if (typeof define === "function" && define.amd) {
        // AMD
        define(factory);
    } else if (typeof exports === "object") {
        // Node, CommonJS-like
        module.exports = factory();
    } else {
        // Browser globals (root is window)
        return factory;
    }
})(this, function() {

    let ts = 0;
    const ns = "_c";
    const unique = {};

    function rgba(opacity = 0.4) {
        // returns String
        //
        // Generates an Array of random rgb colours,
        // formatted CSS value eg. "rgb(123, 23, 111, 0.4)"

        const array = [...Array(4)].map((_, n) => n > 2 ? opacity : (Math.random() * 255) >> 0);
        return `rgba(${array.join(",")})`;
    }

    function getConsoleArguments() {
        const time = (performance.now() >> 0) / 1000;
        const diff = ((Math.max(0, time - ts) * 100) >> 0) / 100;
        const args = arguments && arguments.length ? [...arguments] : ["LOG"];
        const id = args.splice(0, 1)[0];
        const color = unique[id] || (unique[id] = rgba());
        const array = [
            `%c[%s] %s`,
            `background: ${color}; padding: 2px;`,
            `${time >> 0}s (${diff})`,
            id,
            ...args
        ];
        ts = time;
        return array;
    }
    return getConsoleArguments;
    // return window[ns] = getConsoleArguments;
});
