(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['student','models/TabData','device','plugin','modules/debug'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('student'),
        require('models/TabData'),
        require('device'),
        require('plugin'),
        require('modules/debug'),
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(CSParams, TabData, Device, Plugin, debug) {
  // const debug = console.debug;
  class Tab {

    // Tab Polling
      static PollActive()
      {
        // debug("in POLL");
        // _checkInterval(); // checkInterval creates circular reference and doesn't alter flow.
        chrome.tabs.query({ active:true }, this.getDetails);
      }

      pollSelectedTab()
      {
          Tab.PollActive()
      }
      // END Tab Polling

      static Make(target,key)
      {
        // debug('Creating ' + key + ' tab');
        // Returns a promise.
        return CreateTab(chrome.extension.getURL(target), key);
      }

     static MakeExternal(target)
     {
       // debug('Creating ' + target + ' tab');
       // returns a promise.
       return CreateTab(target, target);
     }

    static Update(tabid, changeinfo, tab) {
      if (changeinfo.url)
      {
          // url eg is "http://www.cheese.com/" or "http://www.bbc.co.uk/sport"
          //alert("url updated, url=" + changeinfo.url);
          var redirect = mustFilter(changeinfo.url);
          if (redirect)
          {
            //mustFilter(changeinfo.url); // uncomment to debug mustFilter returning non null
            //alert("redirecting to " + redirect);
            chrome.tabs.update(tabid, { url: redirect });
          }
          else
            chrome.tabs.get(tabid, this.getDetails);
      } else if (changeinfo.title && tab){
          this.getDetails([tab]);
      } else if (changeinfo.status == "complete"){
          // debug("UPD: [COMPLETE]" );
          this.getDetails([tab]);
      }
    }

    static getDetails(tabs)
    {
        if (tabs && tabs[0])
        {
          var tab = tabs[0];
          var tabid = tab.id;

            for (var cid in CSParams.connections)
            {
                if (CSParams.connections[cid].webpoll)
                {
                    var connection = CSParams.connections[cid];
                    // 11/01/19 not sure the line below is used
                    if (!connection.appidlast)
                    {
                        connection.force = true;
                        TabData.send(tab);
                        // this.SendTabData(tab);
                        connection.appidlast = tabid;
                        connection.force = false;
                    }

                    var sameurl = connection.weburllast == tab.url;
                    var sametitle = connection.webtitlelast == tab.title;
                    if (sameurl && sametitle)
                    {
                      // debug("---   " + "["+ tab.id +"] " + tab.title);
                      // same as last time - don't resend
                    }
                    else
                    {
                      connection.weburllast = tab.url;
                      connection.webtitlelast = tab.title;
                    }
                }
            }

            // this.SendTabData(tab);
            TabData.send(tab);
        }
    }

    static getAllTabsFromOpenWindows()
    {
      return new Promise(
        function(resolve) {
        chrome.windows.getAll({ populate: true}, function(w) {
          resolve(w
            .map(browserWindow => browserWindow.tabs)
            .reduce((acc,currentValue) => acc.concat(currentValue)));
        });
      });
    }

    static updateCurrentWebsite(currentWindow)
    {
      return new Promise(function(resolve) {
        if (currentWindow.id > -1){
          chrome.tabs.query({ windowId: currentWindow.id, active:true }, Tab.getDetails);
        }

        resolve(currentWindow);
      });
    }

    static SendData(tab)
    {
      TabData.send(tab);
    }
    // SendTabData(x)
    // {
    //   const plugin = new Plugin();
    //   plugin.appClose
    //
    //
    // }

    static clearId(tabId)
    {
      Plugin.webClose('','',tabId);
      Plugin.appClose(tabId);
      var item = CSParams.TabStore[tabId];
      debug(item +' Tab Removed');
      if (item)
      {
        CSParams.TabStore[item]=false;
        CSParams.TabStore[tabId]=false;
      }
    }

    static Remove(name)
    {
      var found = CSParams.TabStore[name];
      if (found)
      {
        chrome.tabs.remove(CSParams.TabStore[name]);
        if(chrome.runtime.lastError) { return found; }
      }

      return found;
    }

    static replace(addedTabId, removedTabId){
      debug("replaceTab:-[" + removedTabId + "]+[" + addedTabId +"]");
      plugin.appStart(addedTabId);
      this.clearId(removedTabId);
    }

    /**

    webfilter

    **/

    static parseWebFilter(cid, webfilter)
    {
        var rx = webfilter.indexOf('redirect:\r\n');
        var ix = webfilter.indexOf('include:\r\n');
        var ex = webfilter.indexOf('exclude:');

        var redirect = '';
        if (rx >= 0)
        {
            redirect = webfilter.substring(rx + 11, ix - 2);

            if (redirect == 'http://localhost/ApprovedWebList.htm')
                redirect = chrome.extension.getURL('ApprovedWebList.htm');

            if (redirect == 'http://localhost/weblock.htm')
                redirect = chrome.extension.getURL('weblock.html');
        }

        var includes = webfilter.substring(ix + 10, ex).toLowerCase();
        var excludes = webfilter.substring(ex + 10).toLowerCase();
        excludes = excludes.split('\r\n');
        //alert("web filter includes=\n" + includes);
        //alert("web filter excludes=\n" + excludes);
        gRedirect = redirect;

        var connection = CSParams.Connections[cid];
        connection.webflag = webfilter.charCodeAt(0) - 80;
        connection.redirect = redirect;
        connection.includes = includes.split('\r\n');
        connection.excludes = excludes;

        if (tcpConnected){
            gWebflag = webfilter.charCodeAt(0) - 80;
            gRedirect = redirect;
            gIncludes = includes.split('\r\n');
            gExcludes = excludes;
        }

        chrome.tabs.query({}, function(tabs) {
            for (var itab in tabs){
                var tab = tabs[itab];

            // url eg is "http://www.cheese.com/" or "http://www.bbc.co.uk/sport"
                //alert("url updated, url=" + changeinfo.url);
                var redirect = mustFilter(tab.url);
                if (redirect)
                {
                    //mustFilter(changeinfo.url); // uncomment to debug mustFilter returning non null
                    //alert("redirecting to " + redirect);
                    chrome.tabs.update(tab.id, {url: redirect});
                }
            }
        });
    }

    static filterTabs()
    {
        chrome.tabs.query({}, function(tabs) {
            for (var itab in tabs){
                var tab = tabs[itab];
                if (mustFilter(tab.url)) {
                  // TODO::
                    if (gRedirect.indexOf('Approved') > 0 || (CSParams.Connections[0] && CSParams.Connections[0].redirect.indexOf('Approved') > 0 )) {
                        chrome.tabs.update(tab.id, {url: chrome.extension.getURL('ApprovedWebList.htm')});
                    } else {
                        chrome.tabs.update(tab.id, {url: chrome.extension.getURL('weblock.html')});
                    }

                }
            }
        });
    }



    /** Legacy
    function removeSelectedTab()
    {
        chrome.tabs.getSelected(null, function(tab) {
            chrome.tabs.query({}, function(tabs) {
              if (tabs.length == 1)
                  chrome.tabs.update(tab.id, {url: 'chrome://newtab/'});
              else
                  chrome.tabs.remove(tab.id);
            });
        });
    }
    **/

  /**
  * Class Closure
  **/
  }


  /**
   * Internal Methods
   */

   // function _checkInterval()
   // {
   //     if (interval.id && interval.last)
   //     {
   //         var diff = Date.now() - interval.last;
   //         // Difference changed from 10 - 61 seconds as onInterval is Called every 60 Seconds.
   //         if (diff > 61000) {
   //             //alert("setInterval seems broken");
   //             if (CSParams.Opened){
   //                 debug("probably not connected");
   //             }
   //         }
   //     }
   // }

  function CreateTab(target,key)
  {
    return new Promise(function(resolve) {
      // Make a tab
      chrome.tabs.create({ url: target }, function(tab) {
        if (tab)
        {
          // Tab created, log it and resolve - HAPPY!
          CSParams.TabStore[tab.id] = key;
          CSParams.TabStore[key] = tab.id;

          debug(key + ' created, Display PowerMode');
          CSParams.powerMode = 'display';
          chrome.power.requestKeepAwake(CSParams.powerMode);
          debug(key + ' Switch PowerMode system in +90S');
          CSParams.powerMode = 'system';
          window.setTimeout(Device.power.switchMode,90000);
          resolve({key: key, tab: tab});
        } else {
          // Something went wrong, check for error - UNHAPPY
          var le = chrome.runtime.lastError;
          if (le)
          {
            // Message - Do we need to create a window?
            var msg = le.message;
            if (msg == "No current window")
            {
              // Create a window
              chrome.windows.create({ url: target }, function(wnd){
                // Grab active tab details.
                chrome.tabs.query({windowId: wnd.id, active:true}, function(tab) {
                  if (tab && tab[0]){ // = HAPPY :)
                    csParams.TabStore[tab[0].id] = key;
                    csParams.TabStore[key] = tab[0].id;
                    // Power mode to show new tab
                    CSParams.powerMode = 'display';
                    chrome.power.requestKeepAwake(CSParams.powerMode);
                    debug(key + ' Switch PowerMode system in +90S');
                    CSParams.powerMode = 'system';
                    window.setTimeout(Device.power.switchMode,90000);
                    resolve({key: key, tab: tab[0]});
                  }
                });
              });
            } else {
              debug("Unable to create tab: " + target);
            }
          }
        }
      });
    });
  }


  // /**
  // *  WebFilter internals
  // */
  //
  function mustFilter(url)
  {

      var e, domain;
      if (url.substr(0, 7) == 'http://')
          url =  url.substr(7);
      if (url.substr(0, 8) == 'https://')
          url = url.substr(8);

      // don't block approvedweblist.htm survey.html, register.html etc
      if (url.indexOf(chrome.extension.getURL('')) >= 0)
          return null;
      // dont restrict built in chrome pages
       if (url.substr(0, 9) == 'chrome://')
          return null;
      if (url.substr(0, 18) == 'chrome-devtools://')
          return null;

      var retval = null;
      var i;
      for (var cid in CSParams.Connections)
      {
          if (CSParams.Connections[cid].webpoll)
          {
              var connection = CSParams.Connections[cid];

              if (url.substr(0, 6) == 'ftp://' && (connection.webflag & 1) !== 0)
                  return chrome.extension.getURL('weblock.html');


              for (i = 0; i < connection.excludes.length; ++i)
              {
                  //var se = connection.excludes[i]; // for debugging
                  if (connection.excludes[i] == '*')
                  {
                      return connection.redirect;
                  }
                  else if (connection.excludes[i].charAt(0) == '\"')
                  {   // fuzzy
                      e = URLMatches(url, connection.excludes[i]);
                      if (e == 2 || e == 1)
                          return connection.redirect;
                  }
                  else
                  {
                      domain = url.substr(0, url.indexOf('/'));
                      if (connection.excludes[i] == domain)
                          return connection.redirect;
                  }
              }

              retval = connection.redirect;
              for (i = 0; i < connection.includes.length; ++i)
              {
                  //var si = connection.includes[i]; // for debugging
                  if (connection.includes[i] == '*')
                  {
                      retval = null;
                  }
                  else if (connection.includes[i].charAt(0) == '\"')
                  {   // fuzzy
                      e = URLMatches(url, connection.includes[i]);
                      if (e == 2 || e === 0)
                          return connection.redirect;
                      else if (e == 3 || e == 1)
                          retval = null;
                  }
                  else
                  {
                      domain = url.substr(0, url.indexOf('/'));
                      if (connection.includes[i] == domain)
                          retval = null;
                  }
              }
              //return connection.redirect;
          }
      }

      if (tcpConnected){
          if (url.substr(0, 6) == 'ftp://' && (gWebflag & 1) !== 0){
                  return chrome.extension.getURL('weblock.html');
          }

          for (i = 0; i < gExcludes.length; ++i)
          {
              //var se = CSParams.Connections[ic].excludes[i]; // for debugging
              if (gExcludes[i] == '*')
              {
                  return gRedirect;
              }
              else if (gExcludes[i].charAt(0) == '\"')
              {   // fuzzy
                  e = URLMatches(url, gExcludes[i]);
                  if (e == 2)
                      return gRedirect;
              }
              else
              {
                  domain = url.substr(0, url.indexOf('/'));
                  if (gExcludes[i] == domain)
                      return gRedirect;
              }
          }

          retval = gRedirect;
          for (i = 0; i < gIncludes.length; ++i)
          {
              //var si = CSParams.Connections[ic].includes[i]; // for debugging
              if (gIncludes[i] == '*')
              {
                  retval = null;
              }
              else if (gIncludes[i].charAt(0) == '\"')
              {   // fuzzy
                  e = URLMatches(url, gIncludes[i]);
                  if (e == 2)
                      return gRedirect;
                  else if (e == 3)
                      retval = null;
              }
              else
              {
                  domain = url.substr(0, url.indexOf('/'));
                  if (gIncludes[i] == domain)
                      retval = null;
              }
          }
      }

      return retval;
  }

  // // enum (none=-1,nomatch=0,match=1,block=2,allow=3)
  function URLMatches(url, match)
  {
      var q2 = match.indexOf('\"', 1);
      if (q2 > 1)
      {
          var urlpath = url.substr(url.indexOf('/')).toLowerCase(); // skip domain
          var keyword = match.substring(1, q2);
          if (url.indexOf(keyword) >= 0)
          {
              var subUrlScore = 0;
              var hasSubUrl = false;

              while (match.charAt(q2 + 1) == '{' || match.charAt(q2 + 1) == '[')
              {
                  hasSubUrl = true;
                  var localScore = 0;
                  var scoreMult = match.charAt(q2 + 1) == '{' ? 1 : -1;
                  var em = match.indexOf(scoreMult == 1 ? '}' : ']', q2 + 2);
                  var subUrl = match.substring(q2 + 2, em);
                  if (urlpath.substr(0, subUrl.length) == subUrl)
                  {
                      localScore += scoreMult * subUrl.length;
                      if (Math.abs(localScore) > Math.abs(subUrlScore))
                          subUrlScore = localScore;
                  }
                  q2 = em;
              }
              if (hasSubUrl)
              {
                  if (subUrlScore === 0)
                      return 0;
                  return subUrlScore > 0 ? 3 : 2;
              }
              return 1; // was 3; // originally 0
          }
      }
      return -1;
  }

  return Tab;
}));
