define(['lib/batch','logger'], function(batch, log){
  var identity = {
    user: undefined,
    device: {
      long: undefined,
      short: undefined
    },
    userdomain: undefined
  };

  var idBuffer = batch.squash(batch.token.identityAuth, 25000);

  function get(){
    return identity;
  }

  const refresh = {
    user: refreshIdentity,
    device: refreshDevice
  };


  function appender(wall, brick)
  {
    return wall + brick;
  }

  const rGuid = new RegExp(/(.+?)-(.+?)-(.+?)-(.+?)-(.+)/);

  function split(guString)
  {
    return rGuid
      .exec(guString)
      .slice(1)
      .reduce(appender)
      .split(/([^-])./)
      .reduce(appender);
  }

  function refreshIdentity()
  {
      return new Promise(function(resolve, reject) {
        log.all("chrome.identity.getProfileUserInfo");
        chrome.identity.getProfileUserInfo(function(ui){
          if (ui.email === "") {
            idBuffer()
            .then(function(){
                // log.debug("chrome.identity.getAuthToken");
                // chrome.identity.getAuthToken({interactive: true});
                reject("no-identity");
            });
          } else {
            identity.user = ui.email;
            resolve(ui.email);
          }
        });
      });
  }

  function refreshDevice() {
      return new Promise(function(resolve, reject) {
        if (chrome.enterprise && chrome.enterprise.deviceAttributes)
        {
          chrome.enterprise.deviceAttributes.getDirectoryDeviceId(function(devId){
              if (devId){
                identity.device.long = devId;
                identity.device.short = split(devId);
                resolve(identity.device);
              } else {
                identity.userdomain = identity.user;
                resolve(undefined);
              }
          });
      } else {
          identity.userdomain = identity.user;
          resolve(undefined);
      }
    });
  }

  return {
    get: get,
    refresh: refresh
  };
});
