define(['modules/document', 'jquery'], function(document, $){
  ////////////////
  // PUBLIC API //
  ////////////////
  var ko = {
    observable: Observable,
    applyBindings: applyBindings
  };

  ////////////////////
  // PUBLIC METHODS //
  ////////////////////

  // Create an Observable Property
  function Observable(newValue)
  {
    var fn = function(args) {
      switch(arguments.length){
        case 0:
          return fn.currentValue;
        case 1:
          fn.previousValue = fn.currentValue;
          fn.currentValue = args;
          if (fn.onChanged)
          {
            document.dispatchEvent(fn.onChanged);
          }
          break;
        default:
          break;
      }
    };

    fn.onChanged = undefined;
    fn.targets = [];
    fn.currentValue = newValue;
    fn.previousValue = undefined;
    return fn;
  }

  function applyBindings(viewModel)
  {
    if (viewModel.__bound__)
    {
      return false;
    }

    // Map all bound elements
    var boundElements = document.body.querySelectorAll('[data-bind]');
    for (var i = 0; i< boundElements.length; i++){

      // target element
      var targetElement = boundElements[i];
      var data = $(targetElement).data('bind');

      // data-bind element
      if (data){
        // Syntactic variation support

        var bindings = data.replace(/: +/,':').split(' ');
        if (bindings && bindings.length >0)
        {
          for (var b = 0; b < bindings.length; b++)
          {
            var binding = bindings[b].split(':');
            var bindingName, vmData;

            if (binding && binding.length == 2){
              // binding target.
              switch(binding[0]){
                case "value":
                  // bindingName
                  bindingName = binding[1];
                  if (viewModel && viewModel[bindingName])
                  {
                    vmData = viewModel[bindingName];
                    vmData.onChanged = new Event(bindingName);
                    document.addEventListener(bindingName, UpdateTargetValue.bind(vmData));

                    var currentValue = vmData();
                    vmData.targets.push(targetElement);
                    $(targetElement).val(currentValue);
                    $(targetElement).change({
                      model: vmData,
                      targetElement: targetElement
                    }, boundChange);
                  }
                  break;
                case "enable":
                  bindingName = binding[1];
                  if (viewModel && viewModel[bindingName])
                  {
                    vmData = viewModel[bindingName];
                    vmData.onChanged = new Event(bindingName);
                    vmData.targets.push(targetElement);
                    document.addEventListener(bindingName, ShowHideTargets.bind(vmData));
                  }
                  break;
                case "class":
                  bindingName = binding[1];
                  if (viewModel && viewModel[bindingName])
                  {
                    vmData = viewModel[bindingName];
                    vmData.onChanged = new Event(bindingName);
                    vmData.targets.push(targetElement);
                    document.addEventListener(bindingName, classSwitch.bind(vmData));
                  }
                  break;
                case "title":
                  bindingName = binding[1];
                  if (viewModel && viewModel[bindingName])
                  {
                    vmData = viewModel[bindingName];
                    vmData.onChanged = new Event(bindingName);
                    vmData.targets.push(targetElement);
                    document.addEventListener(bindingName, titleChange.bind(vmData));
                  }
                  break;
              default:
                break;
              }
            }
          }
        }
      }
    }

    viewModel.__bound__ = true;
  }

  //////////////////////
  // BINDING HANDLERS //
  //////////////////////

  // Update the Target fields.
  function UpdateTargetValue()
  {
    if (!this.targets)
      return false;

    for (var i=0;i<this.targets.length;i++)
    {
      var identifier = this.targets[i];
      if (identifier)
      {
        var target = $(identifier);
        if (target)
        {
          target.val(this.currentValue);
          // console.log(("updateTargets: " + this.currentValue);
        }
      }
    }
  }

  function ShowHideTargets()
  {
    if (!this.targets)
      return false;

    for (var i=0;i<this.targets.length;i++)
    {
      var identifier = this.targets[i];
      if (identifier)
      {
        if (this.currentValue)
        {
          show(identifier);
        }
        else
        {
          hide(identifier);
        }
      }
    }
  }

    function classSwitch()
    {
      if (!this.targets) {
        return false;
      }

      for (var i=0; i<this.targets.length; i++)
      {
        var identifier = this.targets[i];
        if (identifier)
        {
          if (this.previousValue)
          {
            if (identifier.classList.contains(this.previousValue))
            {
              identifier.classList.remove(this.previousValue);
            }
          }

          identifier.classList.add(this.currentValue);
        }
      }
    }

    function titleChange() {
      if (!this.targets)
        return false;

      for (var i=0;i<this.targets.length;i++)
      {
        var identifier = this.targets[i];
        if (identifier)
        {
          identifier.title = this.currentValue;
        }
      }
    }

  function boundChange(info)
  {
    var data = info.data;
    data.model($(data.targetElement).val());
  }

  function show(element)
  {
    $(element).removeAttr('disabled');
  }

  function hide(element)
  {
    $(element).attr("disabled","disabled");
  }

 // PUBLIC API
  return ko;
});
