console.log("videoStream.js");
(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define([
        'background/thumb/config',
        'lib/batch'
        // 'settings'
    ], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('background/thumb/config'),
        require('lib/batch')
        // require('settings')
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(managedRTC,batch) {
    var canvas = document.createElement('canvas');
    var ctx = canvas.getContext('2d');
    var vid = document.createElement('video');
    var established = undefined;
    var dimensions = { width:undefined, height:undefined };
    var interval = 800;
    var interval_handle = undefined;
    var _stream = undefined;
    var _previous = undefined;
    var _tabImage = { captured: undefined };
    var silent = false;

    var requestBuffer = batch.squash(batch.token.desktopStream, 8000);

    function enable()
    {
      // Established if specified
      established = true;
    }

    function confirmStream(begin)
    {
      return new Promise(function(resolve) {
        if (established) {
          if (!_stream || !_stream.active) {
            // Capture legacy mechanism;
            // chrome.tabs.captureVisibleTab(thumbnail => {
            //   _tabImage.capture = thumbnail
            // });
            requestBuffer()
            .then(initialiseStream)
            .then(s => resolve(s));
          } else {
            getDimensions()
            .then(resolve);
          }
        }
      });
    }

  function initialiseStream() {
    return new Promise(function(resolve) {
      if (!silent) {
        managedRTC.getConfig()
        .then(chooseDesktopMedia)
        .then(k => {
          if (k.video.mandatory.chromeMediaSourceId  =="") {
            return initialiseStream();
          };

          let s = createStream(k)
          .then(stream => {
            _stream = stream;
            _stream.addEventListener('inactive',() => confirmStream(established));
            vid.onloadedmetadata = e => {
              dimensions.width = vid.videoWidth;
              dimensions.height = vid.videoHeight;
              canvas.width = dimensions.width;
              canvas.height = dimensions.height;
              synchronizeInterval(interval);
              resolve(dimensions);
            }

            vid.autoplay = true;
            vid.srcObject = stream;

            vid.onloadeddata = dataloaded;
          });
        });
      }
    });
  }

  function synchronizeInterval(interval)
  {
    clearInterval(interval_handle);
    interval_handle = setInterval(e => drawOnCanvas(ctx, vid), interval)
  }

  function dataloaded(event)
  {
    console.log('Yay! The readyState just increased to  ' +
    'HAVE_CURRENT_DATA or greater for the first time.');
    drawOnCanvas();
    vid.onloadeddata = dataloaded;
  }

  function drawOnCanvas() {
    ctx.drawImage(vid, 0, 0, vid.videoWidth, vid.videoHeight);
  }

  function getCanvas() {
    return canvas;
  }

  function getDimensions(defaultValue = {}) {
    let { width, height } = (dimensions && dimensions.width) ? dimensions : defaultValue;
    return new Promise(function(resolve) {
      if (width) {
        resolve({ width, height });
      }
    });
  }

  function setDisplaySize(width, height) {
    if (width && height) {
      dimensions.width = width;
      dimensions.height = height;
      canvas.width = width;
      canvas.height = height;
    }
  }

  function getVideo() {
    return vid;
  }

  function getStream() {
    return _stream;
  }

  function isActive() {
    return (_stream && _stream.active);
  }

  function Jacques_Merde(silence)
  {
    // Silence!
    silent = (silence) ? true : false;
  }

//   let state = null;
//
//   let moduleName = "webrtc";
//   let area = "managed";
//   // var set = { configuration: undefined };
//
//   // function config(){
//   //   return set.configuration
//   // }
//   //
//   // settings
//   //   .load(moduleName, area)
//   //   .then(data => {
//   //       set.configuration = data;
//   //     });
//   // console.log(managedRTC.config());
//
//   // function chooseDesktopMedia(sourceType) {
  function chooseDesktopMedia(config) {
      // console.log(managedRTC.config());
      // get media dimension etc (aka constraints)
      const { sourceType, constraints } = config;
      // const { sourceType, constraints } = config();

      // shallow copy that is augmented and returned
      let settings = { ...constraints };

      // extension media (aka desktop, screen or tab)
      if (_previous) {
        chrome.desktopCapture.cancelChooseDesktopMedia(_previous);
      }
      return new Promise(resolve => {
          _previous = chrome.desktopCapture.chooseDesktopMedia(sourceType, (streamId, options) => {
              // console.debug("chooseDesktopMedia", sourceType, streamId, options);

              // spread the config constraints values
              // and add chrome specific stuff
              settings = {
                  ...settings,
                  video: {
                      mandatory: {
                          chromeMediaSource: "desktop",
                          chromeMediaSourceId: streamId
                      }
                  },
                  audio: options.canRequestAudioTrack
              };

              resolve(settings);
          });
      });
  }

  function createStream(settings)
  {
    return new Promise((resolve,reject) => {
      navigator.mediaDevices
      .getUserMedia(settings)
      .then(stream => {
//         state.localStream = stream;
        resolve(stream);
      })
      .catch(e => {
        // resetState();
        reject(e);
      });
    });
  }

  function disable()
  {
    established = false;
    if (_previous){
      chrome.desktopCapture.cancelChooseDesktopMedia(_previous);
    }
  }
//
//   // export { getStream };
//   function getStream()
//   {
//     return state.localStream;
//   }
//
//
//   // export { resetState };
//   function resetState(prev) {
//       // check previous state and reset to default
//       // close open cosetDisplaySizennections, streams etc.
//
//       prev = prev || state || null;
//       prev && prev.local ? prev.localstreamer.stream.create.getTracks().forEach(t => t.stop()) : null;
//
//       const zeroState = {
//           started: false,
//           local: null
//       };
//
//       // return a shallow copy
//       return (state = { ...zeroState });
//   }
//
//   return {
//     chooseDesktopMedia,
//     stream: {
//       create: createStream,
//       get: getStream
//     },
//     reset: resetState
//   };
  return { getCanvas, getVideo, getStream, getDimensions, confirmStream, setDisplaySize, enable, disable, isActive, silence: Jacques_Merde};
}));
