(function (factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['student', 'Transmit', 'utf8', 'comms.websocket','showlib','disconnect','models/register','models/Tab','models/sysinfo','modules/debug','modules/mph', 'background/thumb/videoStream','on/reconnect'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('student'),
        require('Transmit'),
        require('utf8'),
        require('comms.websocket'),
        require('showlib'),
        require('disconnect'),
        require('models/register'),
        require('models/Tab'),
        require('models/sysinfo'),
        require('modules/debug'),
        require('modules/mph'),
        require('background/thumb/videoStream'),
        require('on/reconnect')
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(function(CSParams, Transmit, UTF8, Comms, Show, disconnect, Register, Tab, sysinfo, debug, mph, Stream, reconnect) {
  var cachedHelpRequest, notification = undefined;
  var surveyResults = '';

  class exports {
    constructor(target){
      // this = object;

      exports.bind(target)

      return target;
    }

    static bind(target)
    {
      // var API_Methods = this;

      // API Methods
      // exports.sendChatMsg = sendChatMsg;

      /*
      Class and Lesson
       */
      target.getTeacherName        = this.getTeacherName;
      target.getLessonName         = this.getLessonName;
      target.getRoom               = this.getRoom;
      target.getStudent            = this.getStudent;
      target.getUserNameOnly       = this.getUserNameOnly;
      target.studentRegister       = this.studentRegister;
      target.cancelRegister        = this.cancelRegister;


      /*
      Rooms
       */
      target.setRoom               = this.setRoom;
      target.handshake             = this.handshake;
      target.announceRoom          = this.announceRoom;

      /*
      Q & A
       */
      target.sendQAResponse        = this.sendQAResponse;

      /*
      Feedback
       */
      target.sendFeedbackResponse  = this.sendFeedbackResponse;

      /*
      Survey
       */
      target.getSurveyQuestion     = this.getSurveyQuestion;
      target.getSurveyAnswers      = this.getSurveyAnswers;
      target.getSurveyResults      = this.getSurveyResults;
      target.sendSurveyResponse    = this.sendSurveyResponse;
      target.setSurveyResults      = this.setSurveyResults;
      target.cancelSurvey          = this.cancelSurvey;

      /*
      Help Request
      */
      target.sendHelpRequest       = this.sendHelpRequest;


      /*
      Show
      */
      target.scaleShow             = this.scaleShow;
      target.getShow               = this.getShow;
      target.isFullScreenShow      = () => Show.isFullScreen;
      target.getShowDimensions     = this.showDimensions;

      /*
      Notification
      */
      target.showHRNotification    = this.showHRNotification;

      target.getActiveSocket       = this.getActiveSocket;
      target.getSysInfo            = () => sysinfo;
      target.makePPH               = mph.make;
      target.hangupAndCloseConnection = disconnect.reset;
      target.hangupCloseAsync      = disconnect.resetNext;
      target.accelerateReconnect   = this.accelerateReconnect;   
      target.dbg = {
          getCanvas: Stream.getCanvas,
          getVideo: Stream.getVideo,
          getStream: Stream.getStream
        };
      target.gdc = gdc;
    }

    static senddata(cid, d)
    {
        return new Transmit(cid).Send(d);
    }

    static sendImageData(payload)
    {
      return new Transmit(payload.cid).Send(payload.content);
    }
    
    static getTeacherName() {
        return CSParams.Register.Teacher; // "Teacher Name";
    }

    static getLessonName() {
        return CSParams.Register.Lesson; // "Lesson Name";
    }

    static getRoom() {
        return CSParams.Register.Room; // "Room Name";
    }

    static getStudent() {
        var student = CSParams.Register.Student;
        return student ? student : false; // Student Details
    }

    static getUserNameOnly() {
        return CSParams.Register.Flags & 512 ? true : false;
    }

    /**
     * Register
     */
    static studentRegister(student)
    {
      CSParams.Register.student(student)
      .then(function(reg){
        Transmit.senddata(CSParams.Register.Cid, reg);
        disconnect.Register();
      });
    }

    /**
     * Q AND A
     */

    static sendQAResponse(time, answer)
    {
        Transmit.senddata(CSParams.QA.Cid, String.fromCharCode(198, 0, 198, 0, 0, 0, 1, 0, CSParams.QA.RoundNum, 0, 0, 0, time, 0, 0, 0) + answer + String.fromCharCode(0));
    }

    static sendFeedbackResponse(selected)
    {
        Transmit.senddata(CSParams.Feedback.Cid, String.fromCharCode(236,0,0,0,selected,0,0,0));
        // Sent, now close tab.
        Tab.Remove("Feedback");
    }

    /**
     * SURVEY
    */

    static getSurveyQuestion()
    {
        return CSParams.Survey.Question.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
    }

    static getSurveyAnswers() {
        //return "Yes,No,Maybe";
        return CSParams.Survey.Answers.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
    }

    static sendSurveyResponse(response)
    {
        //send Reponse to Survey here
        //nfr alert("Got response "+response);
        var opts = CSParams.Survey.Answers.split(",").filter(function(v) { return v != ''; });
        var res = -1;
        var i;
        for (i = 0; i < opts.length; ++i)
        {
            if (opts[i] == response)
            {
                res = i;
                var vote = String.fromCharCode(148) + z + String.fromCharCode(res) + z + z;
                if (CSParams.Survey.Cid !== '')

                    Transmit.senddata(CSParams.Survey.Cid, vote);
                break;
            }
        }

        //chrome.browserAction.setIcon({path:"School32.png"});
        //chrome.browserAction.setPopup({popup:""});
        //gSurveyCid = "";
        //cancelSurvey();
        Tab.Remove('Survey');
    }

    static cancelSurvey(){
       disconnect.Survey();
    }
    static cancelRegister() {
       disconnect.Register();
    }

    static setSurveyResults(results) {
        surveyResults = results
    }

    static getSurveyResults() {
        return surveyResults;
    }

    static sendHelpRequest(message)
    {
      if (message)
      {
        if (CSParams.Connections.length===0){
          //cache the message
          cachedHelpRequest = message;
        }

        let helpreq = String.fromCharCode(74, 0);
        //helpreq += GetHelpRequestName();onTabChanged // optional?, max 31 chars
        while (helpreq.length < 34) {
            helpreq += z;
        }

        for (var cid in CSParams.Connections)
        {
            Transmit.senddata(cid,helpreq + UTF8.Encode(message, cid) + z);
        }
      } else if (cachedHelpRequest) {
        sendHelpRequest(cachedHelpRequest);
        cachedHelpRequest = undefined;
        return;
      }
    }

    static scaleShow(height, width)
    {
      CSParams.ScreenDimensions.height = height;
      CSParams.ScreenDimensions.width = width;
      Show.Scale(height, width);
    }

    static getShow(viewId, display, fullScreen)
    {
        var currentImageCanvas = Show.displayCanvas;
        Show.syncCanvas(display);

        if ((Show.view.tab|Show.view.window) !== viewId)
        {
            // We have a new view
            if (Show.view.tab)
            {

                // Close old tab, update with new view(tab)Id.
                Show.closeTab();
                Show.view.tab = viewId;

                // draw image from displayCanvas onto backBufferCanvas
                Show.backBufferContext.drawImage(currentImageCanvas,0,0);

            } else if (Show.view.window)
            {
                Show.closeWindow();
            }
        }
        if (!Show.isShowing){
          if (fullScreen)
          {
            Show.closeWindow(viewId);
          } else {
            Show.closeTab(viewId);
          }
        }

        return Show.backBufferCanvas;
        // return display;
    }

    static showDimensions()
    {
      let { Width, Height } = Show;
      return { Width, Height };
    }


    static showHRNotification(image, title, detail)
    {
        var opt = {
            body: detail,
            icon: image
        };

        if (notification){
            notification.close();
        }

       notification = new Notification(title, opt);
    }

    // static setRoom = setRoom;
    static setRoom(roomName) {
        // debug("setRoom: " + roomName);
        if (sessionStorage.room !== roomName) {
            sessionStorage.room = roomName;
            chrome.storage.local.set({ lastRoom: roomName });
            CSParams.reconnectMode.active = 0;
        }

        if (CSParams.Opened) {
            CSParams.Room.Changed = true;
            debug("setRoom (old room name):[ " + CSParams.Room.Name + " ]");
            if (roomName != CSParams.Room.Name) {
                leaveRoom();
                CSParams.Room.Name = roomName;

                Comms.instance.STATUS(roomName, 1);

                if (CSParams.reconnectMode.active) {
                    announceRoom(CSParams.reconnectMode.id);
                }

                announceRoom(roomName);
                if (tcpPage) {
                    tcpPage.postMessage({
                        clientName: localStorage.clientName,
                        macAddress: localStorage.macAddress,
                        roomName: localStorage.room
                    });
                }
            }
        }
    }

    // static handshake = handshake;
    static handshake()
    {
      debug('HANDSHAKE');
      chrome.storage.local.set({ connectionState: "connected_NCS" });
      clearTimeout(CSParams.reconnectMode.expiryHandler);
      CSParams.reconnectMode.expiryHandler = setTimeout(CSParams.reconnectMode.clear, 10000);
      debug('Connected, Open Channel');

      debug('Send OPEN');
      Comms.instance.OPEN();

      //var roomName = (sessionStorage.room) ? sessionStorage.room : '';
      var roomName;
      // CSParams.Room.Name = ;
      if (sessionStorage.room) {
        roomName = sessionStorage.room;
      } else if (localStorage.roomType == 'roomSpecified'){
        roomName = localStorage.room;
        // Set Session room to specified room as this is what was transmitted to the Gateway
        // Query - is this assignment going to cause issues?
        sessionStorage.room = localStorage.room;
      }

      // Reset RoomName as cleared by disconnect.
      debug('STATUS');
      Comms.instance.STATUS(roomName,0);

      debug('ANNOUNCE: ' + roomName);
      if (CSParams.reconnectMode.active)
      {
        // Announce user mode reconnect id
        announceRoom(CSParams.reconnectMode.id);
        // Comms.instance.STATUS(CSParams.reconnectMode.id,0);
      }

      // Announce room mode
      announceRoom(roomName);
      debug('Configure a HeartBeat');
      Comms.mechanism.HeartBeat();
    }

    static accelerateReconnect()
    {
      // Comms.mechanism.accelerateReconnect();
      reconnect.trigger();
      // setTimeout(function(){ document.dispatchEvent(Subscription.on.interval.call); },1500);
    }

    // static announceRoom = announceRoom;
    static announceRoom(room) {
      debug('announce room: ' + room);
      room = room || sessionStorage.room;

      if (CSParams.ServerVersion >= 1.4){
        var utf8BM = 0;
        var utf8stuff = '';

        var ansi = { name: localStorage.clientName, room: room };

        if (UTF8.Required(ansi.name)) {
          utf8BM = 2+1;
          ansi.name = '<name>'; // something safe for use with btoa
        }

        if (UTF8.Required(ansi.room)) {
          utf8BM |= 8+1;
          ansi.room = '<room>'; // something safe for use with btoa
        }

        if (utf8BM !== 0) {
          utf8stuff = String.fromCharCode(utf8BM);

          if (utf8BM & 2) utf8stuff += UTF8.PrefixEncode(localStorage.clientName) + z;

          if (utf8BM & 8) utf8stuff += UTF8.PrefixEncode(room) + z;
        }

        var nc_announce = String.fromCharCode(185,0, 0x0,0x10,0,0, 0x00,0x0a,0,0, 0x5,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0) + ansi.name + z + ansi.room + z + utf8stuff;
        Comms.instance.BROADCASTDATA(room, nc_announce);
      }
    }

    static getActiveSocket(){
      // Comms may be offline.
      return (Comms.instance) ? Comms.instance.websocket : undefined;
    }
  }


  // function sendChatMsg(message) {
  //     message = message;
  //     //send message
  //     //alert("Got message "+message);
  // }



/**
Internals
 */
  function leaveRoom() {
      for (var cid in CSParams.Connections) {
          // CHANNELUPDATE
          new Transmit(cid, String.fromCharCode(188, 0, 0, 0)).Send()
      }
  }

  function gdc(val,sender)
  {
      var evt = new Event(arguments.callee.name);
      crypto.subtle.digest({ name: "SHA-256" }, val).then(function(b){
          var values = new Uint32Array(b);
          var preComp = new Uint32Array(i17n);
          for (var i in values)
          {
              if (values[i] !== preComp[i])
              {
                  return false;
              }
          }
          if (sender.constructor.name == "HTMLInputElement")
          {
              sender.ownerDocument.defaultView.window.dispatchEvent(evt);
          } else {
              sender.dispatchEvent(evt);
          }
      });
  }




  return exports;

  // return { bind } ;
}));
