(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['student','modules/debug','modules/extension','models/Tab','background/thumb/videoStream'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('student'),
        require('modules/debug'),
        require('modules/extension'),
        require('models/Tab'),
        require('background/thumb/videoStream')
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(CSParams, debug, Extension, Tab, Stream) {

  let API = {
    takeScreenshot:takeScreenshot,
    bareScreenshot:bareScreenshot,
    canvas: undefined,
    hiDPICanvas: undefined,
    tabCanvas: document.createElement('canvas'),
    setDisplaySize: setDisplaySize
  };

  function bareScreenshot() {
    return sendScreenshot()
  }

  function takeScreenshot() {
    return new Promise(function(resolve,reject){
      sendScreenshot(true)
      .then(sendImageCanvas)
      .then(resolve);
    });
  }

  function sendScreenshot(withPolling)
  {
    return new Promise(function(resolve, reject) {

      debug("in Take SS");

      if (!API.hiDPICanvas) {
        API.hiDPICanvas = Stream.getCanvas();
      }

      if (API.canvas && !API.context) {
        API.context = API.canvas.getContext('2d');
      }

      if (withPolling) {
        Tab.PollActive();
      }

      if (!Stream.isActive())
      {
        getTabThumbnail()
        .then(drawImage)
        .then(resolve);
      } else {
        resolve(API.hiDPICanvas);
      }
    });
  }

  function sendImageCanvas(hiDPIcanvas)
  {
    return new Promise(function(resolve, reject) {
      for (var cid in CSParams.Connections)
      {
        let connection = CSParams.Connections[cid];
        var { width:thumbw, height:thumbh, thumbnailType, thumblast } = connection;
        API.canvas.width = width;
        API.canvas.height = height;
        API.context.drawImage(hiDPIcanvas,0,0, width, height);
        let smallImage = API.canvas.toDataURL(thumbnailType, 0.7);
        // console.log("Thumbsmall :",width,height);
        if(height < 10) {
          reject("invalidSize");
        }
        // console.log(smallImage);
        if (smallImage != thumblast)
        {
          connection.thumblast = smallImage;
          var comma = smallImage.indexOf(',');
          let payload = {
            cid: cid,
            content: String.fromCharCode(191,0) + atob(smallImage.substr(comma + 1))
          }
          resolve(payload);
        }
        connection.thumbs = 0; // seconds since last sent=0
      }
    });
  }

  function setDisplaySize(width, height) {
    API.tabCanvas.width = width;
    API.tabCanvas.height = height;
    Stream.setDisplaySize(width, height);
  }

  function getTabThumbnail()
  {
    return new Promise(function(resolve, reject) {
      chrome.tabs.captureVisibleTab(null, function(img) {
        var le = chrome.runtime.lastError;
        if (!img) {
          // Default Image
          // img=localStorage.defaultImg;
        }

        if (le && le.message)
        {
          if (le.message == "Failed to capture tab: unknown error")
          {
            // Probably lost focus
          } else if (le.message.search(/(chrome-)*devtools:\/\//) !== -1){
            // devtools (Chrome security policy)
          } else {
            console.debug("unable to capture tab: " + le.message);
          }
        }

        if (img)
        {
          resolve(img);
        }
      });
    });
  }


  // function drawThumbnailCanvas()
  // function registerThumbnailCanv
  //
  function drawImage(bigImage)
  {
      return new Promise((resolve, reject) => {
        img = document.createElement("img");
        img.onload = function(e) {
          drawImageCanvas(e.target)
          .then(resolve);
        };
        img.onerror = reject;
        img.src = bigImage;
        img = null;
      });
  }



  function drawImageCanvas(image)
  {
      return new Promise((resolve, reject) => {
        var { width, height } = API.hiDPICanvas;
        var portContext = API.hiDPICanvas.getContext('2d');
        portContext.fillRect(0,0, width, height);

        var contentRatio = image.width / image.height;
        var portRatio = width / height;
        var scale, origin = undefined;

        let getDrawSize = (scale) => { return { width: image.width * scale, height: image.height * scale }; }

        scale = (contentRatio > portRatio) ?
        // Landscape
        width / image.width :
        // Portrait
        height / image.height;

        let ds = getDrawSize(scale);
        origin = {x: (width - ds.width) / 2, y: (height - ds.height) / 2 };

        portContext.drawImage(image, origin.x, origin.y, ds.width, ds.height);

        resolve(API.hiDPICanvas);
      });
  }

  return API;
}));
