(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define([
        'settings',
        'modules/image',
        'modules/extension',
        'modules/translate',
        'comms.websocket',
        'socket',
        'logger',
        'background/subscription',        
        'background/socketHandlers',
        'background/thumb/videoStream',
      ], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('settings'),
        require('modules/image'),
        require('modules/extension'),
        require('modules/translate'),
        require('comms.websocket'),
        require('socket'),
        require('logger'),
        require('background/subscription'),
        require('background/socketHandlers'),
        require('background/thumb/videoStream'),
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(settings,Image,Extension,Translate,WSComms,Socket,log,Subscription,sock,Stream) {



  function initialize()
  {
    log.debug('Enter setup');

    if (!Image.canvas) {
        Image.canvas = document.createElement('canvas');
        Image.canvas.width = 512;
        Image.canvas.height = 384;
        var context = Image.canvas.getContext('2d');
        var imageObj = document.createElement("img");
        imageObj.onload = function() {
            context.drawImage(this, 0, 0);
            var dataUrl = Image.canvas.toDataURL();
            localStorage.defaultImg = dataUrl.substring(dataUrl.indexOf(',') + 1);

            Image.canvas.width = View.width;
            Image.canvas.height = View.height;
            imageObj.onload = function() {
                context.drawImage(this, 0, 0);
                localStorage.defaultViewImg = Image.canvas.toDataURL();
            };
            imageObj.src = 'images/blankview.png';
        };

        imageObj.src = 'images/default_view_canvas.png';
        Translate();
    }

    if (!Subscription.on.interval.id) {
        Subscription.on.interval.setInterval(60000);
    }

    if (localStorage.pinServerAddress === undefined) {
        localStorage.pinServerAddress = "";
    }

    if (localStorage.pinServerPort === undefined) {
        localStorage.pinServerPort = "";
    }

    NSLibrary.settings.get.managed()
    .then(options => {
      if (options && options.useLegacyThumbnails) {
        Stream.silence(options.useLegacyThumbnails);
      };

      return new Promise(resolve => {
        resolve(options);
      });
    })
    .then(NSLibrary.settings.update.localGateway)
    .then(function(){
        if (WSComms.mechanism)
        {
          var reset = {
            triggered: false
          };
          chrome.storage.local.get("connectionState", function(state){
            if (state == "disconnected" && reset.triggered == false)
            {
              reset.triggered = true;
              WSComms.mechanism.recycle();
            }
          });
          if (WSComms.mechanism.isStale())
          {
            chrome.storage.local.set({connectionState: "disconnected"});
            if (reset.triggered == false) {
              reset.triggered = true;
              WSComms.mechanism.recycle();
            }
          }
        }
        else
        {
          // Comms.mechanism needs to be created.
          console.log('CREATE|Socket|Comms.mechanism');
          new Socket(sock);
        }
    });
  }

  function interval(e)
  {
      debug('interval');
      if (e)
      {
          e.cancelBubble = true;
      }
      if ( !CSParams.Opened || !WSComms.instance || !WSComms.instance.websocket || WSComms.mechanism.isStale()){
        console.log(`Interval - Calling Setup", websocket ${WSComms.instance?WSComms.instance.websocket:'No Instance'}, opened: ${CSParams.Opened}`);
          initialize();
      }

      // API.last = Date.now();

      var takess = false;
      var polltabs = false;
      for (var cid in CSParams.Connections)
      {
          var connection = CSParams.Connections[cid];
          if (connection.thumbi) // thumb interval
          {
              ++connection.thumbs;
              if (connection.thumbs >= connection.thumbi)
              {
                  takess = true;
              }
          }

          if (connection.webpoll)
              polltabs = true;
      }

      if (takess) {
        Stream.confirmStream();
        Image.takeScreenshot()
        .then(Extension.sendImageData);
      }

      //if (polltabs)
        //  pollSelectedTab();

      if (WSComms.instance && WSComms.instance.websocket && CSParams.Opened)
      {
          var room = localStorage.room;
          if (room != CSParams.Room.Name && !CSParams.Room.Changed)
          {
              CSParams.Room.Name = room;
              WSComms.instance.STATUS(room,0);
          }
      }
  }

  new Subscription(Subscription.keys.interval, interval).setInterval(60000);

  return initialize;
}));
