define([
    "background/observable",
    "models/useridentity",
    "background/exportFeature",
    "modules/translate"
], function(
    ko,
    identity,
    ExportFeature,
    Translate
) {
  // Page VM
  function OptionsViewModel(op)
  {
    var self = this;
    self.gatewayAddress = ko.observable(op.gatewayAddress);
    self.gatewayPort = ko.observable(op.gatewayPort);
    self.allowManualClientName = ko.observable(op.allowManualClientName);
    self.allowManualMacName = ko.observable(op.allowManualMacName);
    self.clientName = ko.observable(op.clientName);

    if (op.keyboardLayout)
    {
      op.keyboardLayout = op.keyboardLayout.toLowerCase();
    } else {
      op.keyboardLayout = "en-gb";
    }

    self.keyboardLayout = ko.observable(op.keyboardLayout);
    self.macAddress = ko.observable(op.macAddress);
    self.room = ko.observable(op.roomName);
    self.roomList = ko.observable(op.roomList);
    self.gatewayAddressEnabled = ko.observable(true);
    self.gatewayPortEnabled = ko.observable(true);
    self.keyboardLayoutEnabled = ko.observable(true);
    self.roomEnabled = ko.observable(true);
    self.roomModeEnabled = ko.observable(true);
    self.roomListEnabled = ko.observable(true);
    self.passPhraseEnabled = ko.observable(true);
    return self;
  }

  // Default Values
  var vm = new OptionsViewModel({
    allowManualClientName: null,
    allowManualMacName: null,
    clientName: localStorage.clientName,
    gatewayAddress: (localStorage.gatewayAddress || ''),
    gatewayPort: localStorage.gatewayPort,
    macAddress: (localStorage.macAddress?localStorage.macAddress.substring(0,12):undefined),
    roomName: localStorage.room,
    roomMode: null,
    roomList: (localStorage.roomList || ''),
    keyboardLayout: localStorage.keyboardLayout
  });

  function checkSave(key)
  {
    var el = document.getElementById(key);
    if (el && el.disabled) {
      // Disabled option, do not update
    } else {
      localStorage[key] = vm[key]();
    }
  }

  // Saves options to localStorage.
  function save_options() {
    console.log('save_options')
    checkSave('clientName');
    checkSave('gatewayAddress');
    checkSave('macAddress');
    checkSave('roomList');
    checkSave('gatewayPort');
    checkSave('keyboardLayout');

    var roomName = document.getElementById('roomName');
    if (!roomName.disabled) {
      localStorage.room = vm.room();
    }

    var bgPage = chrome.extension.getBackgroundPage();
    if (localStorage.roomType == 'roomSpecified'){
        bgPage.setRoom(localStorage.room);
    }

    if (($('#ppc')[0].disabled === false) && $('#ppc').val().length > 0){
      chrome.runtime.sendMessage({ updateOptions: { data: $('#pass-output').text() }});
    }

    // Update status to let user know options were saved.
    var status = document.getElementById('status');

    status.innerHTML = Translate('OPTIONS_SAVED');
    setTimeout(function() {
      status.innerHTML = '&nbsp;';
    }, 1900);

    bgPage.hangupCloseAsync()
    .then(bgPage.accelerateReconnect);
  }

  // Restores select box state to saved value from localStorage.
  function restore_options() {
    Translate();

    // ensure single event binding to "click"
    $('#saveBtn').off('click').on('click', function(ev){
      ev.preventDefault();
      save_options();
    });
  }

  function setTextWatcher() {
    $("#macAddress").on('keydown paste', function(e) {
      if (e.type === "keydown") {
        isValidInput(undefined, e);
      } else {
        var clipboardData = e.originalEvent.clipboardData;
        if (clipboardData) {
          isValidInput(clipboardData.getData('text'), e);
        }
      }
    });
  }

  function isValidInput(data, e) {
    var isValid;
    var regex = /^[a-zA-Z0-9\-\:]*$/

    if (data) {
      isValid = regex.test(data);
    } else {
      isValid = regex.test(e.key);
    }

    if (!isValid) {
      $("#macAddress").addClass("red-border");
      e.preventDefault();
      window.setTimeout(function() {
        $("#macAddress").removeClass("red-border");
      }, 500)
    } else {
      $("#macAddress").removeClass("red-border");
    }
  }

  function radioChange() {
    console.log('radioChange');
    localStorage.roomType = this.value;
    if (localStorage.roomType != 'roomSpecified') {
        localStorage.room = '';
        chrome.extension.getBackgroundPage().sessionStorage.removeItem('room');
    }
  }

  function showValues() {
    ko.applyBindings(vm);

    var inputs = document.querySelectorAll('input[name=room]');
    for (var i = 0; i < inputs.length; i++) {
      inputs[i].onchange = radioChange;
    }

    roomType(localStorage.roomType);
  }

  function roomType(value)
  {
    var selector = 'input[name=room][value='+value+']';
    var room = document.querySelector(selector);
    if (room)
    {
        room.click();
        room.checked = true;
    }
  }

  function sendUpdate()
  {
      var textValue = new TextEncoder('utf-8').encode(chrome.runtime.getManifest().name+$('#password')[0].value);
      chrome.extension.getBackgroundPage().gdc(textValue,this);
  }

  function bindToKeyup()
  {
    $('#roomName').on('keyup',function(){
      $('#roomMode').click();
    });
    $('#roomList').on('keyup',function(){
      $('#listMode').click();
    });
  }

  function bindToExportEvents()
  {
    $('#export-as-file').on('click', function(e){
        e.preventDefault();
        $('#download-file').prop('disabled', true);
        $('#file-generation-dialog').css('display', 'block');
    });

    $('#file-generation-dialog, .close-options').on('click', function(e){

      if(e.target.id == 'file-generation-dialog' || e.target.className == 'close-options')
          $('#file-generation-dialog').css('display', 'none');
    });

    $('#active-meta-url').on('change', function(){

      var meta = $('body').find('#meta-details-container');
      if($(this).is(':checked'))
      {
        meta.css('display', 'block');

        if(meta.val().trim() === '')
            $('#generate-file').prop('disabled', true);
        else
            $('#generate-file').prop('disabled', false);
      }
      else{
          meta.css('display', 'none');
          $('#generate-file').prop('disabled', false);
      }
    });

    $('body').on('keyup', '#meta-url', function(){

      if($(this).val().trim() === '')
          $('#generate-file').prop('disabled', true);
      else
          $('#generate-file').prop('disabled', false);
    });

    $('#generate-file').on('click', function(){

      var json = ExportFeature.generateExportFile();

      $('#download-file').prop('disabled', false).find('a').attr('href', 'data:text/plain;charset=utf-8,'+ encodeURIComponent(JSON.stringify(json, null, "\t")));
      document.getElementById('download-link').click();
    });
  }

  function PPH(a)
  {
    $('#pass-output').text(a.detail);
  }

  function makePPH(){
    console.log('makePPH');
    if ($('#ppc')[0].value === ''){
       $('#pass-output').text('');
    } else {
      chrome.extension.getBackgroundPage().makePPH($('#ppc')[0].value,this);
    }
  }

  function setPPH(){
    console.log('setPPH')
    var gd = new Uint8Array(chrome.extension.getBackgroundPage().gd());
    var ot = [];
    for (var a in gd){ot.push(gd[a].toString(16));}
    var p = document.getElementById('password');

    chrome.extension.getBackgroundPage().gdc(new TextEncoder('utf-8').encode(chrome.runtime.getManifest().name+'password'),this);
    p.addEventListener('keyup', sendUpdate, false);
    document.getElementById('ppc').addEventListener('keyup', makePPH, false);

    // document.getElementById('canvas').replaceWith(chrome.extension.getBackgroundPage().getStreamCanvas());
    // document.getElementById('canvas').replaceWith(chrome.extension.getBackgroundPage().getStreamCanvas());
  }

  function setReadOnly(value, element, showTarget)
  {
    if (value)
    {
      element(value);
      showTarget(false);
    } else {
      showTarget(true);
    }
  }

  function applyStoredPolicy(policy) {
    var gws, room;
    if (!policy || !policy.options)
    {
      vm.allowManualClientName(true);
      vm.passPhraseEnabled(true);
      vm.allowManualMacName(true);
      identity.refresh.device()
      .then(function(deviceId)
      {
        if (deviceId){
          setReadOnly(deviceId.short.substring(0,12), vm.macAddress, vm.allowManualMacName);
        }
      });
    } else {
      gws = policy.options.gateways;
      room = policy.options.room;
      vm.allowManualClientName(policy.options.allowManualClientName);
      vm.allowManualMacName(policy.options.allowManualMacName);
      if (policy.options.resetClientName) {
        identity.refresh.user()
        .then(function(userName){
          if (userName)
          {
            var index = userName.indexOf("@");
            if (index)
            {
              vm.clientName(userName.substring(0,index).replace(/\s/g, "").substring(0,30).toUpperCase());
            }
          }
        });
      }

      if (policy.options.useFullDeviceIdentifier)
      {
        identity.refresh.device()
        .then(function(deviceId)
        {
          if (deviceId){
            setReadOnly(deviceId.long, vm.macAddress, vm.allowManualMacName);
          }
        });
      }

      vm.passPhraseEnabled(!policy.options.passKey);
      if (policy.options.keyboardLayout)
      {
        vm.keyboardLayout(policy.options.keyboardLayout.toLowerCase());
        vm.keyboardLayoutEnabled(false);
      } else {
        vm.keyboardLayoutEnabled(true);
      }
    }

    if (gws && gws[0])
    {
      var gw = gws[0];
      setReadOnly(gw.address, vm.gatewayAddress, vm.gatewayAddressEnabled);
      setReadOnly(gw.port, vm.gatewayPort, vm.gatewayPortEnabled);
    } else {
      vm.gatewayAddressEnabled(true);
      vm.gatewayPortEnabled(true);
    }

    vm.roomModeEnabled(true);
    vm.roomEnabled(false);
    vm.roomListEnabled(false);

    if (!room) {
      vm.roomEnabled(true);
      vm.roomListEnabled(true);
    } else {
      if (room.mode) {
        roomType("room"+(room.mode == "Fixed" ? "Specified" : room.mode));
        vm.roomModeEnabled(false);
      }
      vm.room(room.name);
      vm.roomList(room.list);
    }
  }

  function setOp()
  {
    // console.log((e);
    $('.hidden').each(function(){$(this).removeClass('hidden');});
    //$(':disabled').each(function(){$(this)

    chrome.storage.managed.get("options", applyStoredPolicy);

    window.addEventListener('storage', function(e){
      if (e.key === 'clientName') {
        vm.clientName(e.newValue);
      }
    });
    chrome.storage.onChanged.addListener(function(change, area) {
      // console.log(('store: ' + store);
      // console.log(('change: '+ change);
      // console.log(('options: ' + change.options);
      //if (store == 'managed' && change.options && change.options.newValue){
      if (change.options){
        // console.log(('Apply ' + store + ' policy.');
        // console.log((change.options.newValue);
        applyStoredPolicy({options: change.options.newValue});
      }
    });

    $('div#pass').addClass('hidden');
    // Call KO Bind Here
    showValues();
  }

  identity.refresh.device().then(function(deviceId) {
      if (!deviceId) return;
      setReadOnly(deviceId.short.substring(0, 12), vm.macAddress, vm.allowManualMacName);
  });

  // Bind Window Loads

  bindToExportEvents();
  bindToKeyup();
  setPPH();
  setTextWatcher();

  // This ensure translate only extecutes once.
  // NOTE: webRTC is async DOM injection.
  setTimeout(restore_options, 50);

  window.addEventListener('gdc', setOp, false);
  window.addEventListener('make', PPH, false);

  return {};
});
