(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define([], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory();
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function() {
    var includeLabels, dropShadow, data, labels, colors, canvas;

     var sumTo = function(a, i) {
        var sum = 0;
        for (var j = 0; j < i; j++) {
            sum += a[j];
        }
        return sum;
    };

    // helper functions
    var degreesToRadians = function(degrees) {
        return (degrees * Math.PI) / 180;
    };

    var drawSegmentLabel = function(context, startingAngle, arcSize, radius, i) {
        var labelX = canvas.width / 2 + (radius / 2) * Math.cos(startingAngle + arcSize / 2);
        var labelY = canvas.height / 2 + (radius / 2) * Math.sin(startingAngle + arcSize / 2);
        var fontSize = Math.floor(canvas.height / 32);
        context.font = fontSize + 'pt Helvetica';
        context.textAlign = "center";
        context.fillText(labels[i], labelX, labelY);
    }

    var drawSegment = function(canvas, context, i, size, isSelected, includeLabels) {

        context.save();
        var centerX = Math.floor(canvas.width / 2);
        var centerY = Math.floor(canvas.height / 2);
        var radius = Math.floor(canvas.width / 2) - 20;

        var startingAngle = degreesToRadians(sumTo(data, i));
        var arcSize = degreesToRadians(size);
        var endingAngle = startingAngle + arcSize;

        context.beginPath();
        context.moveTo(centerX, centerY);
        context.arc(centerX, centerY, radius, startingAngle, endingAngle, false);
        context.closePath();

        if (isSelected) {
            context.fillStyle = colors[i][1];
        }else {
            context.fillStyle = colors[i][0];
        }

        context.fill();
        context.restore();

        if (includeLabels && (labels.length > i) && size > 0) {
            drawSegmentLabel(context, startingAngle, arcSize, radius, i);
        }
    };
    var draw = function() {

        var context = canvas.getContext('2d');

        if (dropShadow) {
            context.shadowOffsetX = 5;
            context.shadowOffsetY = 5;
            context.shadowBlur = 10;
            context.shadowColor = '#aaaaaa';
            var centerX = Math.floor(canvas.width / 2);
            var centerY = Math.floor(canvas.height / 2);
            var radius = Math.floor(canvas.width / 2) - 21;
            context.beginPath();
            context.moveTo(centerX, centerY);
            context.arc(centerX, centerY, radius, 0, Math.PI * 2, false);
            context.closePath();
            context.fill();
            context.shadowOffsetX = 0;
            context.shadowOffsetY = 0;
            context.shadowBlur = 0;
        }
        for (var i = 0; i < data.length; i++) {
            drawSegment(canvas, context, i, data[i], false, includeLabels);
        }
    };


    PieChart = function(id, o) {
        includeLabels = false;
        if (o.includeLabels === undefined) {
            includeLabels = false;
        }
        else {
            includeLabels = o.includeLabels;
        }
        if (o.dropShadow === undefined) {
            dropShadow = false;
        }
        else {
            dropShadow = o.dropShadow;
        }
        data = o.data; // in degrees
        labels = o.labels ? o.labels : [];
        colors = o.colors ? o.colors : [
                ['#bbddb3', '#1d8e04'], // green
                ['#e2f5b4', '#9edd08'], // yellow green
                ['#fdfbb4', '#faf406'], // yellow
                ['#fbd4b7', '#f2700f'], // orange
                ['#f8bdb4', '#ea2507'], // red
                ['#e2bcbd', '#9e2126']  // purple
        ];
        labels = o.labels ? o.labels : [];
        if (typeof id == 'string') {
            canvas = document.getElementById(id);
        }else {
            canvas = id;
        }

        draw();
    };

    return PieChart;
}));
