(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['modules/document'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('modules/document')
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(document) {
  class plugin {
    constructor() {
    }

    static webStart(url, title, tabid)
    {
      if (!this.isAcceptableChars(title)){
          title = url;
      }

      var web = this.build({ id:2, cmd:1 }, 'dssd', [tabid, url, title, 0]);
      this.sendAsync('sendWebStart', web);
    }

    static webClose(url, title, tabid)
    {
      if (!this.isAcceptableChars(title)){
          title = url;
      }

      var web = this.build({ id:2, cmd:2 }, 'dss', [tabid, url, title]);
      this.sendAsync('sendWebClose', web);
    }

    static appStart(tabid)
    {
      var appid = tabid;
      var title = 'chrome.exe';
      var app = this.build({ id:1, cmd:1 }, 'dds', [appid, tabid, title]);
      this.sendAsync('sendAppStart', app);
    }

    static appClose(tabid)
    {
      var appid = tabid;
      var app = this.build({ id:1, cmd:2 }, 'dd', [appid, tabid]);
      this.sendAsync('sendAppClose', app);
    }

    static appChange(url, title, tabid)
    {
      var appid = tabid;

      if (!this.isAcceptableChars(title)){
          title = url;
      }

      var app = this.build({ id:1, cmd:4 }, 'dds', [appid, tabid, title]);
      this.sendAsync('sendAppChange', app);
    }

    static appActivate(tabid)
    {
      var appid = tabid;
      var app = this.build({ id:1, cmd:3 }, 'dd', [appid, tabid]);
      this.sendAsync('sendAppActivate', app);
    }

    static sendAsync(key, value)
    {
      var pluginEvent = new CustomEvent('pluginEvent', { detail: {'key':key, 'value': value } });
      document.dispatchEvent(pluginEvent);
    }

    static isAcceptableChars(stringToCheck)
    {
      var largeCharFound;
      for (var i in stringToCheck)
      {
        largeCharFound = Math.max(stringToCheck.charCodeAt(i))>255;
        if (largeCharFound) break;
      }

      return !largeCharFound;
    }

    static build(target, format, params)
    {
      // debug("PLUGIN: " + JSON.stringify(target));
      // debug("PLUGIN: " + format + " " + JSON.stringify(params));
      const keys = {
        ID:0,
        FLAGS:1,
        CMD:2,
        PLUGINCMD:3,
        SIZE:4,
        OFFSET:5,
        PKTSIZE:6
      };

      if (!target.id) { return; } // invalid type

      // build parameter block as a string (before building header which needs length of param block)
      var len;
      var pstr = format + z + String.fromCharCode(target.cmd, 0, 0, 0);
      for (var i = 0; i < format.length; ++i)
      {
        switch (format[i])
        {
          case 'd':
          pstr += String.fromCharCode(4, 0, 0, 0) +
          String.fromCharCode(params[i] & 0xFF, (params[i]>>>8) & 0xFF, (params[i]>>>16) & 0xFF, 0);
          break;
          case 's':
          len = params[i].length + 1;
          pstr += String.fromCharCode(len & 0xFF, (len>>>8) & 0xFF, (len>>>16) & 0xFF, 0) +
          params[i] + z;
          break;
          case 'p':
          pstr += z4;
          break;
        }
      }

      pstr += z4; // for safety I think

      var plen = pstr.length; // only the length of params! (not entire packet)

      var head = new ArrayBuffer(32);
      var buf = new Uint32Array(head,4,7);

      // HEADER
      // Plugin (16-Bit)
      new Uint16Array(head,2,1)[0]=150; //plugin;
      buf[keys.ID]=target.id; // Type ({webls:2,app:1})
      buf[keys.FLAGS]=3;
      buf[keys.CMD]=7;
      buf[keys.PLUGINCMD]=target.cmd;
      buf[keys.SIZE]=plen;
      buf[keys.OFFSET]=0;
      // Size (16-Bit)
      buf[keys.PKTSIZE]=plen;

      var out = [String.fromCharCode.apply(null, new Uint8Array(head.slice(2,30)))];

      return out.join('') + pstr;
    }
  }


  return plugin;
}));
