(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory();
  } else {
      // Browser globals (root is window)
      root.NSLibrary = root.NSLibrary || {};
      root.NSLibrary.PubSub = factory();

      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function() {
  let PubSub = {
    topicList: {},
    topicLookup:{},
    subID:0,


    subscribe: function (topic, callback, context) {
      //if the topic doesn't exist, create the topic
      if ( !(this.topicList[topic]) ) {
        this.topicList[topic] = {};
      }

      //add the callback function to the topic under the key subID
      this.subID += 1;
      context = context || this;

      this.topicList[topic][this.subID] = {
        callback: callback,
        context: context
      };

      this.topicLookup[this.subID] = topic;
      return this.subID;
    },

    publish: function (topic) {
      var key, sub, arg = null;

      if (typeof topic !== 'string'){
        throw new Error('Topic must be a string');
      }

      if (arguments[1] !== undefined) {
        arg = arguments[1];
      }

      //check if the topic exists. a topic exists if there are any subscriptions to it
      if (this.topicList[topic]) {
        /* iterate over the subscriptions for this topic and invoke each cb function
        * passing it any arguments from the arguments array at position 1
        */
        for (key in this.topicList[topic]) {
          if (this.topicList[topic].hasOwnProperty(key)) {
            sub = this.topicList[topic][key];
            sub.callback.call(sub.context, arg);
          }
        }
      }
    },

    unsubscribe: function (subID) {
      var topic;

      if (typeof subID !== 'number') {
        throw new Error('Subscription ID of the wrong type.');
      }

      if (this.topicLookup.hasOwnProperty(subID)) {
        topic = this.topicLookup[subID];

        delete this.topicList[topic][subID];
        delete this.topicLookup[subID];
        return true;
      }

      return false;
    }

  };

  return PubSub;
}));
