(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // NSLibrary.Comms
      // AMD
      define(['comms','lib/batch','background/storage','models/useridentity','background/thumb/videoStream','pubsub','modules/debug'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
          require('comms'),
          require('lib/batch'),
          require('background/storage'),
          require('models/useridentity'),
          require('background/thumb/videoStream'),
          require('pubsub'),
          require('modules/debug'),
      );
  } else {
      // Browser globals (root is window)
      root.NSLibrary = root.NSLibrary || { content: {} };
      let NSLibrary = root.NSLibrary;

      // Ensure NSLibrary framework items exist.
      ["squash"].forEach(key => {
        NSLibrary[key] = NSLibrary[key] || {};
        return;
      });

      // NSLibrary.Comms.WebSockets = NSLibrary.Comms.WebSockets || {};

      NSLibrary.settings = {...NSLibrary.settings, ...factory.call(
        NSLibrary.Comms,
        NSLibrary.squash,
        require('background/storage'),
        require('models/useridentity'),
        require('background/thumb/videoStream'),
        require('pubsub')
      )};
  }
}(this, function(Comms, sq, storage, identity, pubsub, Stream, debug) {

  let settings = {};
  var globalStore = { area: undefined };

  function storageChanged(changes, location)
  {
    debug('Storage event received from: '+location);
    debug(JSON.stringify(changes));

    if (location == 'local') {
      if (changes.passKey | changes.activeMetaURL) {

        // Enforce Managed PolicyR
        if (chrome.storage.managed)
        {
          chrome.storage.managed.get(['options'], function(item){
            if (item.options && (item.options.passKey || item.options.activeMetaURL)){
              debug('Storage Event Managed Option Update');
              UpdateMetaOptions({passKey: item.options.passKey, activeMetaURL: item.options.activeMetaURL});
              // This one is going to cause issues!!
              if (item.options.activeMetaURL)
              {
                chrome.tabs.create({url: (item.options.activeMetaURL)});
              }

            } else {
              debug('Storage Event Local - MetaURL / passKey');
              //debug(JSON.stringify(changes));
              UpdateMetaOptions({ passKey: changes.passKey, activeMetaURL: changes.activeMetaURL });
            }
          });
        } else {
          debug('Storage Event, No managed storage.');
          //debug(JSON.stringify(changes));
          UpdateMetaOptions({ passKey: changes.passKey, activeMetaURL: changes.activeMetaURL });
        }
      }

      if (changes.gatewayAddress || changes.gatewayPort || changes.room || changes.roomList || changes.roomType) {
        // Gateway changes, recycle active WebSocket to flush updates.
        // Schedule Recycle (Squash inbound requests to 3500ms and recycle socket once only.

          // Create squasher with parameters
          var squash = sq.squash(sq.token.recycleSocket, 3500);

          // Delay the invocation by squashDuration
          squash.expire()
          .then(squash)
          .then(function(){
            if (Comms.WebSockets && Comms.WebSockets.active) {
              Comms.WebSockets.active.recycle();
            }
          });
        }
    }
    else if (location == 'managed')
    {
        // changes.options.newValue will not be present if config is being removed
        if (changes.options && changes.options.newValue)
        {
          var options = changes.options.newValue;
          chrome.runtime.sendMessage({ updateOptions: { data: options.passKey }});
          // ONLY RESPECT IF CHANGE CAME FROM MANAGED STORAGE.
          if (options.passKey === '' || !options.passKey){
            if (!options.passKey && changes.options.oldValue && changes.options.oldValue.passKey) {
              settings.update.pfxClear();
            } else if (options.passKey === '' && changes.options.oldValue && changes.options.oldValue.passKey !== '') {
              settings.update.pfxClear();
            }
          }

          if (options.resetClientName) {
            storage.handlers.set('clientName', function(e){
              // Former NCR getUserDetails();
            });

            // Former NCR getUserDetails();
          }

          identity.refresh.device()
          .then(function(deviceId)
          {
            if (deviceId){
              if (options.useFullDeviceIdentifier)
              {
                localStorage.macAddress = deviceId.long;
              } else {
                localStorage.macAddress = deviceId.short.substring(0,12);
              }
            }
          });

          chrome.storage.local.set({
            activeMetaURL: options.activeMetaURL
          });

          if (options.activeMetaURL)
          {
            chrome.tabs.create({url: options.activeMetaURL});
          }

          Stream.silence(options.useLegacyThumbnails);

          localStorage.activeMetaURL = options.activeMetaURL;
          localStorage.allowManualClientName = changes.options.newValue.allowManualClientName;

          // gateway will not be present if config is being removed
          var targetJSON = {};
          if (changes.options.newValue.gateways)
          {
            var gw = changes.options.newValue.gateways;
            if (gw && gw[0])
            {
              targetJSON.gatewayAddress = gw[0].address;
              targetJSON.gatewayPort = gw[0].port;
            }
          }

          var room = changes.options.newValue.room;
          if (room)
          {
            targetJSON.room = room.name;
            targetJSON.roomList = room.List;
            targetJSON.roomType = "room" + (room.mode == 'Fixed') ? "Specified" : room.mode;
          }
          setManagedRoom(changes.options.newValue);

          chrome.storage.local.set(targetJSON);
        } else {
          chrome.storage.local.set({
            gatewayAddress: '',
            gatewayPort:  443
          });
        }

        if (changes.debugDest) {
          localStorage.debugDest = changes.debugDest.newValue;
        }
      }
    }

    function runtimeMessage(request, sender)
    {
      //TODO: ADD SPECIFIED ROOM LIST PER CLIENT
      if (request.updateConnectionDetails)
      {
        updateConnectionDetails(request.updateConnectionDetails.data, request.updateOptions)
        .then(ReRunSetup);
      } else if (request.updateOptions) {
        updateOptions(request.updateOptions.data);
      }

      if (request.updatePinDetails)
      updatePinDetails(request.updatePinDetails.data);

      if (request.closeTab)
      chrome.tabs.remove(sender.tab.id);

      if (request.keyPress)
      //TODO SEND CHARACTERS
      console.log(request.data);

      if (request.windowResize)
      pubsub.publish('windowResize');
    }

    function ConvertMetaConnectionDetails(meta){
      return new Promise(function(resolve){
        // Promise (success, failure)
        debug('Convert meta');
        var output = {  };
        var data = meta.connectionDetails;

        if (data){
          // Start
          output.gateways = [];

          if (data.indexOf('/') >-1) {
            var roomParts = data.split('/');

            if (roomParts[1].indexOf(",")>-1){
              //multiple rooms in a list
              output.room = { mode: "List", list: roomParts[1] };
            }else{
              //single room
              output.room = { mode: "Fixed", name: roomParts[1] };
            }
          }

          var gateway = { address: undefined, port: 443 };

          if (data.indexOf(':') < 0) {
            if (data.indexOf('/') >-1) {
              gateway.address = data.substring(0,data.indexOf('/'));
            }else{
              gateway.address = data;
            }

            gateway.port = 443;
          }else {
            var parts = data.split(':');
            gateway.address = parts[0];
            if (data.indexOf('/') >-1) {
              gateway.port = 1 * parts[1].substring(0,parts[1].indexOf('/'));
            }else{
              gateway.port = 1 * parts[1];
            }
          }

          output.gateways[0] = gateway;

        }


        if (meta.updateOptions)
        {
          output.passKey = meta.passKey;
        }

        resolve(output);
      });
    }

    function compareManagedScope(data){
      return new Promise(function(resolve){
        // Promise (success, failure)
        if (data && data.local && data.managed){
          var local = data.local;
          var managed = data.managed;

          var assign = function(key)
          {
            if (local[key] && managed[key])
            {
              local[key] = managed[key];
            }
          };

          assign("gateways");
          assign("room");
          assign("passKey");

          resolve(local);
        } else {
          resolve(data);
        }
      });
    }

    function checkSave(key)
    {
      var el = document.getElementById(key);
      if (el && el.disabled) {
        // Disabled option, do not update
      } else {
        localStorage[key] = el.value;
      }
    }

    function saveOption(item)
    {
      return new Promise(function(resolve){
        // Promise (success, failure)
        if (item && item.key){
          // Save any value if specified - including undefined.
          localStorage[item.key] = item.value;
          resolve(item.value);
        }
      });
    }

    // Saves options to localStorage.
    function save_options(){
      var roomName = document.getElementById('roomName');
      if (!roomName.disabled) {
        localStorage.room = roomName.value;
      }

      var bgPage = chrome.extension.getBackgroundPage();
      if (localStorage.roomType == 'roomSpecified'){
        bgPage.setRoom(localStorage.room);
      }
    }

    function apply(settings){
      return new Promise(function(resolve){
        // Promise (success, failure)
        if (settings)
        {
          if (settings.gateways && settings.gateways[0])
          {
            chrome.storage.local.set({ "gateways": settings.gateways });
            var gw = settings.gateways[0];
            saveOption({ key: 'gatewayAddress', value: gw.address});
            saveOption({ key: 'gatewayPort', value: gw.port});
          }

          if (settings.room)
          {
            // save room details
            chrome.storage.local.set({ "room": settings.room });
            setManagedRoom(settings);
          }

          if (settings.passKey)
          {
            updateOptions(settings.passKey);
          }
        }

        resolve(settings);
      });
    }

    function updateConnectionDetails(data, updateOptions)
    {
      debug('incoming request - updateConnection');

      return new Promise(function(resolve,reject){
        // Promise (success, failure)

        var managedSettings = loadManagedSettings();
        var meta = { connectionDetails: data, updateOptions: updateOptions };
        if (updateOptions)
        meta.passKey = updateOptions.data;

        var metaSettings = ConvertMetaConnectionDetails(meta);

        Promise.all([managedSettings, metaSettings])
        .then(function(d){
          var managed = d[0];
          var metaSettings = d[1];

          if (managed){
            compareManagedScope({ local: metaSettings, managed: managed })
            .then(apply)
            .then(function(){
              resolve(managed);
            });
          } else {
            apply(metaSettings)
            .then(function(){
              resolve(metaSettings);
            });
          }
        }).catch(function(error){
          debug("updateConnectionDetails Failure" + error);
          reject("updateConnectionDetails Failure" + error);
        });
      });
    }

    function updatePinDetails(data)
    {
      debug('incoming request - updatePin');
      if (data.indexOf(':') < 0) {
        localStorage.pinServerAddress = data;
        localStorage.pinServerPort = 443;
      } else {
        var parts2 = data.split(':');
        localStorage.pinServerAddress = parts2[0];
        localStorage.pinServerPort = parts2[1];
      }
    }

    function updateOptions(data, sender)
    {
      debug('incoming request - updateOptions');

      var p = new Promise(function(resolve,reject){
        chrome.storage.managed.get('options', function(managed){
          if (managed && managed.options && managed.options.passKey)
          {
            data = managed.options.passKey;
          }

          if (data && typeof(data) === "string") {
            resolve(data.match(/[0-9a-f]{2}/g));
          }
        });
      }).then(function(options){
        var oa = new Uint8Array(i17n);

        for (var i in options)
        {
          oa[i] = parseInt(options[i],16);
        }

        if (sender && sender.url) {
          if (sender.url.replace(/chrome-extension:\/\/.+\/options.html/g,'').length !== 0){
            if (chrome.storage.local){
              chrome.storage.local.get('i17n', function(s)
              {
                if (s.i17n)
                {
                  chrome.tabs.create({url: s.i17n});
                }
              });
            }
          } else {
            chrome.storage.local.set({i17n: sender.url});
          }
        }

        chrome.storage.local.set({passKey: data});
      });
    }

    function pfxClear(){
      //debug(JSON.stringify(options));
      var pfx = $(new TextEncoder('utf-8').encode(chrome.runtime.getManifest().name));
      var ab = new ArrayBuffer(8+ pfx.length);
      var u8 = new Uint8Array(ab);
      u8.set(pfx);
      var dv = new DataView(ab);
      dv.setUint32(pfx.length, 1936941424, true);
      dv.setUint32(4 + pfx.length, 1685221239, true);

      crypto.subtle.digest({name:"SHA-256"}, ab)
      .then(function(dg)
      {
        var a8 = new Uint8Array(dg);
        var op = [];
        for (var i in a8){
          op.push(('0'+a8[i].toString(16)).slice(-2));
        }

        NSLibrary.settings.update.options(op.join(''));
      });
    }

    function setManagedRoom(store){
      var r = store.room;
      if (r && r.mode)
      {
        switch(r.mode) {
          case "Fixed":
            localStorage.roomType = "roomSpecified";
            localStorage.room = r.name;
            chrome.storage.local.set({
              roomType: "roomSpecified",
              room: r.name
            });
            chrome.extension.getBackgroundPage().setRoom(r.name);
            break;
          case "List":
            localStorage.roomList = r.list;
            localStorage.roomType = "roomList";
            chrome.storage.local.set({
              roomType: "roomList",
              roomList: r.list
            });
            break;
          case "Manual":
            localStorage.roomType = "roomManual";
            chrome.storage.local.set({
              roomType: "roomManual"
            });
            break;
          default:
        }
      }
    }

    function setManagedGatewayAndRoom(){
      return new Promise(function(resolve){
        // Promise (success, failure)
        chrome.storage.managed.get("options", function(store){

          if (store.options)
          {
            var gw = store.options.gateways;
            if (gw && gw[0])
            {
              localStorage.gatewayAddress = gw[0].address;
              localStorage.gatewayPort = gw[0].port;
            }

            setManagedRoom(store.options);
          }

          // resolve gateway
          resolve({address: localStorage.gatewayAddress, port:localStorage.gatewayPort});
        });
      });
    }

    function setGateway(managedOptions){

      return new Promise(function(resolve){
        // Promise (success, failure)
        if (managedOptions){
          debug('loadManagedStore - Connection Info');

          if (managedOptions.options) {
            // Sets local values for RoomType and Name, or List.
            NSLibrary.settings.setManagedRoom(managedOptions.options);

            if (managedOptions.options.gateways && managedOptions.options.gateways[0]) {
              // Pull Managed / Local / Defaults
              localStorage.gatewayAddress = managedOptions.options.gateways[0].address;
              localStorage.gatewayPort = managedOptions.options.gateways[0].port || 443;
            }
          } else {
            localStorage.gatewayAddress = localStorage.gatewayAddress;
            localStorage.gatewayPort = localStorage.gatewayPort || 443;
          }

          resolve({ address: localStorage.gatewayAddress, port: localStorage.gatewayPort });
        } else {
          debug('managed storage not found');
          gateway = { address: localStorage.gatewayAddress, port: localStorage.gatewayPort };

          if (!gateway.port)
          {
            gateway.port = 443;
          }

          localStorage.gatewayAddress = gateway.address;
          localStorage.gatewayPort = gateway.port;

          resolve (gateway);
        }
      });
    }

    function loadLocalSettings(){
      return new Promise(function(resolve,reject){
        try {
          chrome.storage.local.get(['passKey','activeMetaURL'], function(item){
            debug('loadChromeStorageLocal');
            resolve(item);
          });
        } catch (e){
          reject(e);
        }
      });
    }

    function loadManagedSettings(key = 'options'){
      return new Promise(function(resolve,reject){
        try {
          if (chrome.storage.managed){
            chrome.storage.managed.get([key], function(item){
              debug('loadChromeStorageManaged');
              if (item && item[key])
              {
                resolve(item[key]);
              }

              resolve(undefined);
            });
          } else {
            // no managed storage available
            resolve(undefined);
          }
        } catch (e) {
          reject(e);
        }
      });
    }


    function setStorageArea(key) {
      // returns undefined
      //
      // Sets global variable governing the storage area.
      // A "managed" area will clear any "local" persisted values

      chrome.storage.managed.get(key, data => {
        if (data && data[key]) {
          console.log("Using MANAGED storage.");
          globalStore.area = "managed";

          console.warn("Clearing LOCAL data.");
          chrome.storage.local.clear();
          return;
        }

        chrome.storage.local.get(key, data => {
          console.log("Using LOCAL storage.");
          globalStore.area = "local";
          if (data && !data[key]) {
            // console.warn("No [%s] data exists. Using defaults", key);
          }
        });
      });
    }

    function loadSettings(key, area = null) {
      // returns Promise
      //
      // @key     String      "managed" or "local"
      // @area    String      (optional)
      //
      // the callback resolves to the storage @area
      // and returns data matching the propery of @key

      area = area || globalStore.area || null;
      const illegal = !/^(managed|local)$/.test(`${area}`);

      return new Promise((resolve,reject) => {

        if (illegal) reject("Illegal storage area");

        console.log("Settings loading ...[%s] [%s]", key, area);
        chrome.storage[area].get(key, settings => {
          console.log("Settings found ...");
          resolve(settings[key]);
        });
      });
    }

    setStorageArea("webrtc");

    // Public API
    settings.handleChange = storageChanged;
    settings.runtimeMessage = runtimeMessage;

    settings.load = loadSettings;
    settings.get = {
      local: loadLocalSettings,
      managed: loadManagedSettings
    };

    settings.update = {
      connection: updateConnectionDetails,
      pin: updatePinDetails,
      options: updateOptions,
      localGateway: setGateway,
      pfxClear: pfxClear
    };

    settings.setManagedRoom = setManagedRoom;
    settings.setManagedGatewayAndRoom = setManagedGatewayAndRoom;

    return settings;

}));


// (function(settings){  })(NSLibrary.settings);
