
(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['student','modules/document','modules/debug'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('student'),
        require('modules/document'),
        require('modules/debug')
      );
  } else {
    // TODO:

    // static decodeBMP256
    // static decodeBMP

      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
      // var NSLibrary = NSLibrary || {};
      // NSLibrary.Client = NSLibrary.Client || {};
      // NSLibrary.Client.Show = NSLibrary.Client.Show || {};

  }
}(this, function(CSParams, debug) {
  function logRuntimeError(message){
    return function(){
      var err = chrome.runtime.lastError;
      if (err && err.message){
        console.debug(message + ' - ' + err.message);
      }
    };
  }

  class show
  {
    constructor(){
      this.tempShowCanvas = document.createElement('canvas');
      this.tempShowContext = this.tempShowCanvas.getContext('2d');
      this.backBufferCanvas = document.createElement('canvas');
      this.backBufferContext = this.backBufferCanvas.getContext('2d');
      this.displayCanvas = undefined;
      this.displayContext = undefined;

      this.palette = [];
      this.isShowing = 0;
      this.scaleToFit = 1;
      this.isAnnotating = 0;
      this.annotateCanvas = document.createElement('canvas');
      this.annotateContext = this.annotateCanvas.getContext('2d');
      this.isUpsideDown = false;
      this.lastTab = undefined;
      this.isFullScreen = 1;
      this.view = {tab:undefined, window:undefined};
      this.showInterval = undefined;
      this.xOffset = 0;
      this.yOffset = 0;
      this.fragments = [];
      this.showHold = null;
    }

    syncCanvas(displayCanvas){
      displayCanvas.width = this.Width || this.backBufferCanvas.width;
      displayCanvas.height = this.Height || this.backBufferCanvas.height;
      if(this.backBufferCanvas)
      {
        displayCanvas.getContext('2d').drawImage(this.backBufferCanvas,0,0);
      }
      //
      // this.backBufferCanvas.Width = this.Width;
      // this.backBufferCanvas.Height = this.Height;

      console.log("Sync Canvas, Show:[", this.Width, this.Height, "]");
      console.log("Sync Canvas, display:[", displayCanvas.width, displayCanvas.height, "]");
      console.log("Sync Canvas, BBC:[", this.backBufferCanvas.width, this.backBufferCanvas.height, "]");

      if(this.displayCanvas)
      {
        // Increase size of bbc
        console.log("Sync Canvas, SDC:[", this.displayCanvas.width, this.displayCanvas.height, "]");
        this.backBufferContext.drawImage(this.displayCanvas,0,0);
        displayCanvas.getContext('2d').drawImage(this.backBufferCanvas,0,0);
      }
      //else {
      //}

      displayCanvas.style.webkitTransform = displayCanvas.style.webkitTransform + ";translate3d(0,0,0)";

      this.displayCanvas = displayCanvas;
      this.displayContext = displayCanvas.getContext('2d');
      this.Scale();
    }

    closeTab(id){
      if (!id){
        id = this.view.tab;
        this.view.tab = undefined;
      }
      chrome.tabs.remove(id, logRuntimeError('Error while loading show tab: ' + id));
    }

    closeWindow(id){
      if (!id){
        window.clearInterval(this.showInterval);
        id = this.view.window;
        this.view.window = undefined;
      }
      chrome.windows.remove(id, logRuntimeError('Error while removing window: ' + id));
    }

    endShow(){
        if (this.isShowing===1) {
            CSParams.powerMode = "system";

            this.showHold = window.setTimeout(function(){
                chrome.power.releaseKeepAwake();
                chrome.power.requestKeepAwake(CSParams.powerMode);
            },300000);

            if (this.view.tab!==undefined){
                this.closeTab();
            }
            if (this.view.window!==undefined){
                this.closeWindow();
            }

            this.isShowing = 0;
            if (this.lastTab)
            {
                chrome.tabs.update(this.lastTab,{active:true}, logRuntimeError("Could not activate previous tab: " + this.lastTab));
            }
            this.backBufferCanvas = document.createElement("canvas");
            this.backBufferContext = this.backBufferCanvas.getContext('2d');
        }
    }

    decodeBMP256(bmp) {
        //takes a netSupport bitmap object and convert into an imageData structure

        // context = show.tempShowContext
        // var cnv = context.canvas;
        // var imgData = context.createImageData(cnv.width, cnv.height);
        var imgData = this.tempShowContext.createImageData(this.tempShowCanvas.width, this.tempShowCanvas.height);
        var outCount = 0;
        var char1;
        var bmpLength = bmp.length;

        for (var i = 0; i < bmpLength; i++) {
            char1 = bmp.charCodeAt(i);

            imgData.data[outCount]     = this.palette[char1][0]; // red
            imgData.data[outCount + 1] = this.palette[char1][1]; //green
            imgData.data[outCount + 2] = this.palette[char1][2]; //blue
            imgData.data[outCount + 3] = 255;
            outCount += 4;
        }

        return imgData;
    }

    decodeBMP(bmp, upsideDown) {
        // context = show.tempShowContext
        upsideDown = upsideDown || 0;
        //takes a netSupport bitmap object and convert into an imageData structure

        // var cnv = context.canvas;
        var cnv = this.tempShowCanvas;
        // var imgData = context.createImageData(cnv.width, cnv.height);
        var imgData = this.tempShowContext.createImageData(cnv.width, cnv.height);
        var outCount = 0;
        var i;
        var char1;
        var char2;
        var bmpLength = bmp.length;
        // check if upside down (like normal thumbnails (not chrome specific thumbs))
        if (upsideDown){
            var lineIndex = ((cnv.width * cnv.height - cnv.width) * 4);
            outCount = lineIndex;

            for ( i = 0; i < bmpLength; i += 2) {
                char1 = bmp[i];
                char2 = bmp[i + 1];

                imgData.data[outCount] = (char2 << 1) & 248; // red
                imgData.data[outCount + 1] = ((char2 & 3) << 6) | ((char1 & 224) >> 2); //green
                imgData.data[outCount + 2] = (char1 << 3) & 248; // blue
                imgData.data[outCount + 3] = 255;

                if (i % (cnv.width*2) === 0){
                  lineIndex -= cnv.width * 4;
                  outCount = lineIndex;
                } else{
                  outCount += 4;
                }
            }
        }else{
            for (i = 0; i < bmpLength; i += 2) {
                char1 = bmp.charCodeAt(i);
                char2 = bmp.charCodeAt(i + 1);

                imgData.data[outCount] = (char2 << 1) & 248; // red
                imgData.data[outCount + 1] = ((char2 & 3) << 6) | ((char1 & 224) >> 2); //green
                imgData.data[outCount + 2] = (char1 << 3) & 248; // blue
                imgData.data[outCount + 3] = 255;
                outCount += 4;
            }
        }

        return imgData;
    }

    static strRepeat(pattern, count) {
      if (count < 1) return '';
      var result = '';
      while (count > 0) {
        if (count & 1)
        {
          result += pattern;
        }
        count >>= 1;
        pattern += pattern;
      }
      return result;
    }

    rleDecode(encoded) {
      var bmpData = '';
      var dataLength;
      var toAdd;
      var i = 0, count;
      var runlen = 1;

      if (encoded.charCodeAt(0) == 128) {
        runlen = encoded.charCodeAt(1);
        // runlen should now be 4 for 16bit colour, 2 for 8bit colour
        i = 2;
      }

      while (i < encoded.length) {
        dataLength = encoded.charCodeAt(i);
        if (dataLength > 128) {
          count = dataLength - 128;
          toAdd = encoded.substr(i + 1, runlen);
          bmpData += show.strRepeat(toAdd, count);
          i += runlen + 1;

        }else {
          bmpData += encoded.substr(i + 1, dataLength);
          i += dataLength + 1;
        }
      }

      return (bmpData);
    }

    enforceShow(){
      if (this.view.tab || this.view.window){
        if (this.view.window){
          var w = this.view.window;
          chrome.windows.get(this.view.window, {}, wnd => {
            logRuntimeError('Error while loading show window: ' + w)();

            if (wnd) {
              chrome.windows.update(this.view.window, {focused: true, state: 'fullscreen'});
            }else {
              this.endShow();
              this.startShow();
            }
          });
        }

        if (this.view.tab){
          var t = this.view.tab;
          chrome.tabs.get(this.view.tab, tab => {
            logRuntimeError('Error while loading show tab ' + t)();

            if (!tab) {
              this.isShowing=0;
              this.view.tab = undefined;
              window.clearInterval(this.showInterval);
              this.startShow();
            }
          });
        }
      }
    }

    startShow(){
        //this.backBufferCanvas.getContext('2d').clearRect(0, 0, this.backBufferCanvas.width, this.backBufferCanvas.height);
        if (!this.isShowing) {
         chrome.tabs.query({ currentWindow: true, active: true }, t => {
                this.lastTab = undefined;
                if (t[0])
                {
                    // browser was operating in 'Tabbed' mode (It is possible for t[0] to be undefined)
                    this.lastTab = t[0].id;
                }

                if (this.showHold){
                    window.clearTimeout(this.showHold);
                }

                CSParams.powerMode = 'display';
                chrome.power.requestKeepAwake(CSParams.powerMode);

                if (this.isFullScreen){
                    chrome.windows.create({url: 'show.html', top: 0, left: 0, type: 'popup', width: screen.width, height: screen.height}, wnd => {
                         this.view.window = wnd.id;
                         // From Chrome 38
                         // chrome.tabs.setZoom(wnd.tabs[0].id, 1.00);
                    });

                }else{
                    chrome.tabs.create({url: 'show.html'}, tab => {
                        if (tab)
                        {
                            this.view.tab = tab.id;
                        }
                        else {
                            if(chrome.runtime.lastError)
                            {
                                var message = chrome.runtime.lastError.message;
                                if (message && message == "No current window")
                                {
                                    // Make a window then create a tab.
                                    chrome.windows.create({url: 'show.html', top: 0, left: 0, type: 'normal', width: screen.width, height: screen.height}, wnd => {
                                        // Grab active tab details for show.
                                        chrome.tabs.query({windowId: wnd.id, active:true}, (tab) => {
                                            if (tab && tab[0]){
                                                this.view.tab = tab[0].id;
                                            }
                                        });
                                    });
                                }
                            }
                        }
                    });
                }
             
                // Don't send thumbnails during the show
                window.clearInterval(thumbnailTimer);
                this.showInterval = window.setInterval(() => this.enforceShow.call(this), 500);
                this.isShowing = 1;

            });
        }
    }

    Scale(windowHeight, windowWidth){
        var widthRatio = Math.min(1, windowWidth/ CSParams.Show.Width);
        var heightRatio = Math.min(1, windowHeight / CSParams.Show.Height);

        if (this.scaleToFit)
        {
            CSParams.get.currentWindow()
            .then(w => {

                var windowZoom = Math.min(w.height / windowHeight, w.width / windowWidth);
                // var scale = (this.isFullScreen) ? Math.max(widthRatio, heightRatio) * windowZoom : Math.min(widthRatio, heightRatio);
                // console.log('WZ:'+windowZoom +' Min' +  Math.min(widthRatio, heightRatio) +' Max' +  Math.max(widthRatio, heightRatio) + 'MScale:' + Math.max(widthRatio, heightRatio) * windowZoom);
                var scale = (this.isFullScreen) ? Math.max(widthRatio, heightRatio) * windowZoom : Math.min(widthRatio, heightRatio);
                // console.log('scale::'+scale);
                // this.backBufferCanvas.style.webkitTransformOrigin = "50% 50%";
                this.displayCanvas.style.webkitTransformOrigin = "50% 50%";

                if (windowZoom >=1.06)
                {
                    //console.log('big');
                    // Page bigger than viewport
                    var srw = Math.min(1, w.width/ CSParams.Show.Width);
                    var srh = Math.min(1, w.height / CSParams.Show.Height);
                    scale = Math.max(srw, srh);// * windowZoom;
                    // this.backBufferCanvas.style.webkitTransform = "scale(" + scale + ")";
                    this.displayCanvas.style.webkitTransform = "scale(" + scale + ")";
                }

                if (windowZoom >0.95 && windowZoom < 1.06)
                {
                    // console.log('goldilocks');
                    // Page bigger than viewport
                    scale = Math.min(widthRatio, heightRatio);
                    this.displayCanvas.style.webkitTransform = "scale(" + scale + ")";
                }

                if (windowZoom <=0.95)
                {
                    // console.log('small');
                    // viewport larger than page
                    this.displayCanvas.style.webkitTransform = "scale(" + scale + ")";
                }
            });
        }
        else
        {
            this.displayCanvas.style.webkitTransform = "scale(1)";
        }

        console.log("scale: [", this.Width, this.Height, "]");

        //if(this.Width && this.displayCanvas.height)
        if(this.displayCanvas && this.Width)
        {
          // We have a displayCanvas, and a dimension.
          // this.displayCanvas.style = this.backBufferCanvas.style;

          if(this.expandCanvas.call(this.displayCanvas,this.Width, this.Height))
          {
            this.displayContext.drawImage(this.backBufferCanvas, 0, 0);
          }

          if(this.expandCanvas.call(this.backBufferContext,this.Width, this.Height))
          {
            this.backBufferContext.drawImage(this.displayCanvas, 0, 0);
          }
          // this.displayContext.drawImage(this.backBufferCanvas, 0, 0);
        }
    }

    PushFragment(xpos, ypos, imagedata){
      var frag = {
        xpos: xpos,
        ypos: ypos,
        imagedata: imagedata
      };

      this.fragments.push(frag);
    }

    // * @param  {bool} migrateCanvas Migrate the contents of canvas onto resized canvas.
    /**
    * Check and expand the canvas if the size has changed.
    * @param  {number} width  target width.
    * @param  {number} height target height.
    */
    expandCanvas(width, height)
    {
      if(this.height == height && this.width == width) {
        return false;
      }

      this.height = height;
      this.width = width;
      return true;
    }
  }
  // }

  const instance = new show();
  return instance;
}));
