(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
      // AMD
      define(['settings','student','disconnect','modules/window', 'modules/extension', 'comms.websocket', 'background/subscription','background/platform', 'on/lifecycle'], factory);
  } else if (typeof exports === 'object') {
      // Node, CommonJS-like
      module.exports = factory(
        require('settings'),
        require('student'),
        require('disconnect'),
        require('modules/window'),
        require('modules/extension'),
        require('comms.websocket'),
        require('background/subscription'),
        require('background/platform'),
        require('on/lifecycle'),
      );
  } else {
      // Browser globals (root is window)
      // root.ns = root.ns||{ content: {} };
      // root.ns.content. = factory(root.ns.content.);
  }
}(this, function(Settings, CSParams, disconnect, window, Extension, WSComms, Subscription, platform, lifecycle) {
  var instance = {
    heartBeat: undefined,
    rapidResponder: undefined
  };

  class Socket {
    constructor(handlers){
      if (!lifecycle) {
        require(['on/lifecycle'], l => lifecycle = l);
      }
      this.handlers = handlers;
      this._expired = [];
      this._enabled = false;
      platform.isChromeOS()
      .then(() => this._enabled = true)
      .catch(err => {
        this.enabled = false;
        console.log('Platform Unknown', err);
      });
      this.Create();
    }

    Create(){
      var old = this.socket;
      // var self = this;
      debug('Create Socket');

      return new Promise((resolve,reject) => {
        Settings.setManagedGatewayAndRoom()
        .then(gateway => {
          if (!this._enabled){
            return false;
          }
          if (gateway.address !== '' && gateway.address.toLowerCase() !== 'undefined')
          {
            try {
              window.chrome.storage.local.set({ connectionState: "connecting" });
              this.socket = new WebSocket('ws://' + gateway.address + ':' + gateway.port);
              // Register new Command Handler(mechanism, websocket)
              new WSComms.Commands(this, this.socket);
              Socket.bind(this.handlers, this.socket);
              resolve(this.socket);
            } catch(error) {
              debug("WebSocket create failure: " + error);
              reject(error);
            }
          } else {
            //
            debug('No Socket Address');
            // reject
          }
        });

        if (old)
        {
          if (old.readyState == 1)
          {
            old.close(1000, 'SOCKET_EXPIRED');
          }
          if (old.readyState == 2)
          {
            old.onclose = undefined;
          }
        }

        // TODO: Remove the expired logic, this is probably not needed once the functionality is established.
        this._expired.push(old);
      });
    }

    isStale()
    {
      console.log('stale check..');
      let lc = lifecycle.check();
      if (!this.socket || (this.socket && this.socket.readyState >1) || !lc)
      {
        debug(`is stale - socket:${this.socket}, readyState:${this.socket.readyState}, lifecycle: ${lc}`);
        window.chrome.storage.local.set({ connectionState: "disconnected" });
        return true;
      }
      console.log('not stale');
      return false;
    }

    recycle()
    {
      debug('recycle.socket');
      this.Create();
      // .then(function(sock){
      //   debug('socket recycled');
      //   sock.addEventListener('open', function(event){
      //     if (sessionStorage.room)
      //     {
      //       setRoom(sessionStorage.room);
      //     }
      //
      //     if (CSParams.reconnectMode.active)
      //     {
      //       console.debug('reconnection triggered');
      //       announceRoom(CSParams.reconnectMode.id);
      //     }
      //
      //     if (sessionStorage.room) {
      //       announceRoom(sessionStorage.room);
      //     }
      //   });
      // });
    }

    HeartBeat()
    {
      /**
       * Register Heartbeat event for this socket.
       */
      if (instance.heartBeat)
      {
        debug('RegisterHeartBeat already subscribed.');
        window.clearInterval(instance.heartBeat);
      }

      instance.heartBeat = window.setInterval(() => Subscription.on.heartbeat.call(), 60000);
      debug('Registered HeartBeat ' + instance.heartBeat);
    }


    static bind(handlers, skt){
      skt.onerror = handlers.error;
      skt.onopen = handlers.open;
      skt.onclose = handlers.closed;
      skt.onmessage = handlers.message;
      skt.binaryType = "arraybuffer";

      return skt;
    }

    static accelerateReconnect()
    {
        debug("...Accelerate Reconnect");
        Subscription.on.heartbeat.call();

        var slowReconnect = setTimeout(function(){
            chrome.storage.local.set({ connectionState: "disconnected" });
            disconnect.reset();
            Subscription.on.interval.call();
        }, 7000);

        // Send Poll
        var emitter = setInterval(function(){
          var ts=(function(){ var d = new Date(); return d.toISOString(); })();
          if (WSComms.instance.websocket && WSComms.instance.websocket.readyState === 1)
          {
            WSComms.instance.POLL(ts);
          }
        }, 750);
        debug("Emitting...");

        // Response Received
        function responder(){
            chrome.storage.local.set({ connectionState: "connected_NCS" });
            // clearTimeout(CSParams.reconnectMode.expiryHandler);
            // CSParams.reconnectMode.expiryHandler = setTimeout(clearReconnectMode, 10000);
            debug('...Echo Received');
            debug('Remove Emitter');
            clearInterval(emitter);
            removeEventListener(instance.rapidResponder, responder);
            clearTimeout(slowReconnect);
            // clearTimeout(lifecycle.openHandshake);
            Extension.handshake();
        }

        // Attach Listener
        addEventListener(instance.rapidResponder, responder);

        // Timeout
        setTimeout(function(){
            clearInterval(emitter);
        }, 10000);
    }
  }

  return Socket;
}));
