var NSLibrary = NSLibrary || {};
NSLibrary.Client = NSLibrary.Client || {};
NSLibrary.Client.View = NSLibrary.Client.View || {};

(function(view) {
    "use strict";

    view.canvas = document.createElement('canvas');
    view.context = view.canvas.getContext('2d');
    view.oldImage = document.createElement("img");
    view.newImage = document.createElement("img");
    view.pixelCount = 0;
    view.timeout = 500;
    view.gridSizeX = 0;
    view.GridSizeY = 0;
    view.nrX = 0;
    view.nrY = 0;
    view.timer = 0;
    view.isScraping = 0;
    view.width = 0;
    view.height = 0;
    view.isRestart = 0;
    view.FRAMESTART = String.fromCharCode(189, 0, 1);
    view.FRAMEEND = String.fromCharCode(189, 0, 0);
    view.viewCid = -1;



view.newImage.onload = function() {
            window.clearTimeout(view.timer);
            var data, data2, BMPData, xpos = 0, ypos, encoded, y, oldGridSize, oldPixelCount;

            // iterate through all the parts of the image

            for (y = 0; y < view.nrY; y++) {

                ypos = y * view.gridSizeY;

                if (view.height - ypos < view.gridSizeY) {
                    oldGridSize = view.gridSizeY;
                    view.gridSizeY = view.height - ypos;
                    oldPixelCount = view.pixelCount;
                    view.pixelCount = view.gridSizeY * view.gridSizeX * 4;
                    view.canvas.height = view.gridSizeY;
                }

                view.context.drawImage(view.oldImage, xpos, ypos, view.gridSizeX, view.gridSizeY, 0, 0, view.gridSizeX, view.gridSizeY);
                data = view.context.getImageData(0, 0, view.gridSizeX, view.gridSizeY);


                view.context.drawImage(view.newImage, xpos, ypos, view.gridSizeX, view.gridSizeY, 0, 0, view.gridSizeX, view.gridSizeY);
                data2 = view.context.getImageData(0, 0, view.gridSizeX, view.gridSizeY);

                if (!view.compareImages(data, data2, view.pixelCount) || view.oldImage.src==="") {

                    BMPData = view.createBMP(data2);
                    encoded = view.rleEncode(BMPData);
                    view.sendFrame(encoded, xpos, ypos, view.gridSizeX, view.gridSizeY);
                }

            }
            if (oldGridSize) {
                view.gridSizeY = oldGridSize;
                oldGridSize = undefined;
                view.canvas.height = view.gridSizeY;
            }
            if (oldPixelCount) {
                view.pixelCount = oldPixelCount;
                oldPixelCount = undefined;
            }

            view.oldImage.src = view.newImage.src;
            if (view.isScraping) {
                view.timer = window.setTimeout(view.getScreen, view.timeout);
            }

        };


// takes an imagedata object as argument
   view.createBMP = function(oData) {
    var iWidth = oData.width;
    var iHeight = oData.height;
    var iPadding = (4 - ((iWidth * 3) % 4)) % 4;
    var paddingChar = String.fromCharCode(0);
    var aImgData = oData.data;

    var iOffsetY, iOffsetX, offSet;
    var strPixelRow = '';
    var strPixelData = '';
    var r, g, b;
    var y = iHeight;
    do {
      iOffsetY = iWidth * (y - 1) * 4;
      strPixelRow = '';
      for (var x = 0; x < iWidth; x++) {
        iOffsetX = 4 * x;

        offSet = iOffsetY + iOffsetX;
        r = aImgData[offSet];
        g = aImgData[offSet + 1];
        b = aImgData[offSet + 2];

        strPixelRow += String.fromCharCode((((g >>> 3) << 5) | (b >>> 3)) & 255);
        strPixelRow += String.fromCharCode((((r >>> 3) << 2) | (g >>> 6)) & 255);

      }
      for (var c = 0; c < iPadding; c++) {
        strPixelRow += paddingChar;
      }
      strPixelData += strPixelRow;
    } while (--y);


    return strPixelData;
    };

view.getScreen = function()
{

    require(['modules/image','background/thumb/videoStream'], function(NSImage, videoStream){
      videoStream.confirmStream();
      var img;
      NSImage.bareScreenshot()
      .then(canvas => {
        window.clearTimeout(view.timer);

        img = canvas.toDataURL("image/octet-stream", 0.7) || localStorage.defaultViewImg;;
        view.newImage.src = img;

        if (view.isScraping) {
          view.timer = window.setTimeout(view.getScreen, view.timeout);
        }
      })
    var img = NSImage.hiDPICanvas.toDataURL

    //chrome.tabs.captureVisibleTab(function(img) {
    //    if (!img) {
    //        var le = chrome.runtime.lastError;
    //        if (le && le.message)
    //        {
    //            if (le.message == "Failed to capture tab: unknown error")
    //            {
    //                // Probably lost focus
    //            } else if (le.message.search('chrome-devtools://') !== -1){
    //                // devtools (Chrome security policy)
    //            } else {
    //                debug("unable to capture tab: " + le.message);
    //            }
    //        }

    //        // Don't update image, Reset for next interval.
    //        window.clearTimeout(view.timer);
    //        view.timer = window.setTimeout(view.getScreen, view.timeout);
    //        return;
    //    }

       window.clearTimeout(view.timer);

       //the onload handler does all the work here, but is defined once as a global, instead of each iteration of captureVisibleTab();
       if (view.isScraping) {
         view.timer = window.setTimeout(view.getScreen, view.timeout);
       }
    // });
    });

    //CapturevisibleTab only works if screen has changed. So this is fix if you stop viewing and restart
    if (view.isViewRestart) {
        var tempScreen = view.oldImage.src;
        view.oldImage.src = '';
        view.newImage.src = tempScreen;
    }
};

view.rleEncode = function(data)
{
    var result;
    var i = 0;
    var runlen;
    var remaining = data.length;

    result = String.fromCharCode(128, 4);

    while (remaining !== 0)

    {
        // count number of consecutively equal 4 byte blocks, but no more than 127 of them
        runlen = 0;

        while (runlen < 126 && remaining >= 8 && data.substr(i, 4) === data.substr(i + 4, 4)) {

            ++runlen;
            i += 4;
            remaining -= 4;

        }

        if (runlen) {

            ++runlen;
            result += String.fromCharCode(runlen + 128) + data.substr(i, 4);

            i += 4;
            remaining -= 4;

        } else { // non run (or near the end)

            while (runlen < 124 && remaining >= 8 && data.substr(i + runlen, 4) !== data.substr(i + runlen + 4, 4)) {
                runlen += 4;
                remaining -= 4;
            }

            if (remaining < 8 && runlen <= 120) {
                runlen += remaining;
                remaining = 0;
            }

            result += String.fromCharCode(runlen) + data.substr(i, runlen);
            i += runlen;

        }

    }

    return result;
};

require(['modules/extension'], function(extension){
  view.sendFrame = function (encodedBMPData, xpos, ypos, width, height)
  {

    // debug("in SendFrame");
    // xpos,ypos is top,left of tile in pixels
    // width of tile in pixels
    // height of tile in pixels

    extension.senddata(view.viewCid, view.FRAMESTART);
    var scrape = String.fromCharCode(158,0,(xpos & 255),((xpos >> 8) & 255),(ypos & 255),((ypos >> 8) & 255),(width & 255), ((width >> 8) & 255),(height & 255),((height >> 8) & 255),16) + encodedBMPData;
    extension.senddata(view.viewCid, scrape);
    extension.senddata(view.viewCid, view.FRAMEEND);
  };
})

 view.compareImages = function(img1, img2, pixelCount) {
   for (var i = 0; i < pixelCount; ++i) {
      if (img1.data[i] !== img2.data[i]) {

           return false;
        }
   }
   return true;
};


})(NSLibrary.Client.View);
